from beziers.point import Point

class BoundingBox:
  """A representation of a rectangle within the Beziers world,
  used to store bounding boxes."""

  def __init__(self):
    self.bl = None
    self.tr = None

  def __str__(self):
    return "BB[%s -> %s]" % (self.bl,self.tr)

  """Determine the bounding box - returns the bounding box itself."""
  def bounds(self):
    return self

  @property
  def area(self):
    """Returns the area of the bounding box."""
    vec = self.tr-self.bl
    return vec.x * vec.y

  @property
  def left(self):
    """Returns the X coordinate of the left edge of the box."""
    return self.bl.x

  @property
  def right(self):
    """Returns the X coordinate of the right edge of the box."""
    return self.tr.x

  @property
  def top(self):
    """Returns the Y coordinate of the top edge of the box."""
    return self.tr.y

  @property
  def bottom(self):
    """Returns the Y coordinate of the bottom edge of the box."""
    return self.bl.y

  @property
  def width(self):
    """Returns the width of the box."""
    return self.tr.x - self.bl.x

  @property
  def height(self):
    """Returns the height of the box."""
    return self.tr.y - self.bl.y

  @property
  def centroid(self):
    """Returns a `Point` representing the centroid of the box."""
    return Point((self.left + self.right)*0.5, (self.top+self.bottom)*0.5)

  def extend(self,other):
    """Add an object to the bounding box. Object can be a `Point`,
    another `BoundingBox`, or something which has a `bounds()` method."""
    if isinstance(other, Point):
      if not(self.bl): self.bl = other.clone()
      if not(self.tr): self.tr = other.clone()
      if other.x < self.bl.x: self.bl.x = other.x
      if other.y < self.bl.y: self.bl.y = other.y
      if other.x > self.tr.x: self.tr.x = other.x
      if other.y > self.tr.y: self.tr.y = other.y
    elif isinstance(other, BoundingBox):
      self.extend(other.bl)
      self.extend(other.tr)
    else:
      # Try getting its bb
      self.extend(other.bounds())

  def translated(self, point):
    """Returns a new BoundingBox translated by the vector"""
    bb2 = BoundingBox()
    bb2.bl = self.bl + point
    bb2.tr = self.tr + point
    return bb2

  def includes(self, point):
    """Returns True if the point is included in this bounding box."""
    return self.bl.x >= point.x and self.tr.x <= point.x and self.bl.y >= point.y and self.tr.y <= point.y

  def overlaps(self,other):
    """Returns True if the given bounding box overlaps with this bounding box."""
    if other.left > self.right:
      return False
    if other.right < self.left:
      return False
    if other.bottom > self.top:
      return False
    if other.top < self.bottom:
      return False
    return True

  def addMargin(self,size):
    """Adds a few units of margin around the edges of the bounding box."""
    self.bl = self.bl + Point(-size,-size)
    self.tr = self.tr + Point(size,size)




import sys
from beziers.path.representations.Segment import SegmentRepresentation
from beziers.utils.intersectionsmixin import Intersection
import logging
import pyclipper
from beziers.line import Line
from beziers.point import Point

class BooleanOperationsMixin:

  def getSelfIntersections(self):
    """Returns a list of a path's self-intersections as Intersection objects."""

    segs = self.asSegments()
    intersections = []
    for seg in segs:
      l = seg.hasLoop
      if l and l[0]>0 and l[0]<1 and l[1]>0 and l[0]<1:
        intersections.append(Intersection(seg,l[0], seg,l[1]))
    for i1 in range(0, len(segs)):
      for i2 in range (i1+1, len(segs)):
        for i in segs[i1].intersections(segs[i2]):
          if i.t1 > 1e-2 and i.t1 < 1-1e-2:
            intersections.append(i)
    return intersections

  def removeOverlap(self):
    """Resolves a path's self-intersections by 'walking around the outside'."""
    if not self.closed:
      raise "Can only remove overlap on closed paths"
    splitlist = []
    splitpoints = {}
    def roundoff(point):
      return (int(point.x*1),int(point.y*1))

    for i in self.getSelfIntersections():
      splitlist.append((i.seg1,i.t1))
      splitlist.append((i.seg2,i.t2))
      splitpoints[roundoff(i.point)] = {"in":[], "out": []}
    self.splitAtPoints(splitlist)
    # Trace path
    segs = self.asSegments()
    for i in range(0,len(segs)):
      seg = segs[i]
      if i < len(segs)-1:
        seg.next = segs[i+1]
      else:
        seg.next = segs[0]
      seg.visited = False
      segWinding = self.windingNumberOfPoint(seg.pointAtTime(0.5))
      seg.windingNumber = segWinding
      if roundoff(seg.end) in splitpoints:
        splitpoints[roundoff(seg.end)]["in"].append(seg)
      if roundoff(seg.start) in splitpoints:
        splitpoints[roundoff(seg.start)]["out"].append(seg)
    newsegs = []
    copying = True
    logging.debug("Split points:", splitpoints)
    seg = segs[0]
    while not seg.visited:
      logging.debug("Starting at %s, visiting %s" % (seg.start, seg))
      newsegs.append(seg)
      seg.visited = True
      if roundoff(seg.end) in splitpoints and len(splitpoints[roundoff(seg.end)]["out"]) > 0:
        logging.debug("\nI am at %s and have a decision: " % seg.end)
        inAngle = seg.tangentAtTime(1).angle
        logging.debug("My angle is %s" % inAngle)
        # logging.debug("Options are: ")
        # for s in splitpoints[roundoff(seg.end)]["out"]:
          # logging.debug(s.end, s.tangentAtTime(0).angle, self.windingNumberOfPoint(s.pointAtTime(0.5)))
        # Filter out the inside points
        splitpoints[roundoff(seg.end)]["out"] = [ o for o in splitpoints[roundoff(seg.end)]["out"] if o.windingNumber < 2]
        splitpoints[roundoff(seg.end)]["out"].sort(key = lambda x: x.tangentAtTime(0).angle-inAngle)
        seg = splitpoints[roundoff(seg.end)]["out"].pop(-1)
        # seg = seg.next
        # logging.debug("I chose %s\n" % seg)
      else:
        seg = seg.next

    self.activeRepresentation = SegmentRepresentation(self,newsegs)

  def clip(self,clip,cliptype, flat=False):
    splitlist1 = []
    splitlist2 = []
    intersections = {}
    cloned = self.clone()
    clip = clip.clone()

    # Split all segments at intersections
    for s1 in self.asSegments():
      for s2 in clip.asSegments():
        for i in s1.intersections(s2):
          if i.t1 > 1e-8 and i.t1 < 1-1e-8:
            if i.seg1 == s1:
              splitlist1.append((i.seg1,i.t1))
              splitlist2.append((i.seg2,i.t2))
            else:
              splitlist2.append((i.seg1,i.t1))
              splitlist1.append((i.seg2,i.t2))
            intersections[i.point] = i

    logging.debug("Split list: %s" % splitlist1)
    logging.debug("Split list 2: %s" % splitlist2)
    cloned.splitAtPoints(splitlist1)
    clip.splitAtPoints(splitlist2)
    logging.debug("Self:")
    logging.debug(cloned.asSegments())
    logging.debug("Clip:")
    logging.debug(clip.asSegments())

    segs1unflattened = cloned.asSegments()
    segs2unflattened = clip.asSegments()

    # Replace with flattened versions, building a dictionary of originals
    segs1 = []
    reconstructionLUT = {}
    precision = 100.

    def fillLUT(flats):
      for line in flats:
        key = ((line.start * precision).rounded(), (line.end * precision).rounded())
        reconstructionLUT[key] = (line._orig or line)
        key2 = ((line.end * precision).rounded(), (line.start * precision).rounded())
        reconstructionLUT[key2] = (line._orig or line).reversed()

    for s in segs1unflattened:
      flats = s.flatten(2)
      fillLUT(flats)
      segs1.extend(flats)

    segs2 = []
    for s in segs2unflattened:
      flats = s.flatten(2)
      fillLUT(flats)
      segs2.extend(flats)

    # Leave it to the professionals
    subj = [(s[0].x*precision, s[0].y*precision) for s in segs1]
    clip = [(s[0].x*precision, s[0].y*precision) for s in segs2]
    pc = pyclipper.Pyclipper()
    pc.AddPath(clip, pyclipper.PT_CLIP, True)
    pc.AddPath(subj, pyclipper.PT_SUBJECT, True)
    paths = pc.Execute(cliptype, pyclipper.PFT_EVENODD, pyclipper.PFT_EVENODD)
    outpaths = []

    # Now reconstruct Bezier segments from flattened paths
    def pairwise(points):
      a = (p for p in points)
      b = (p for p in points)
      next(b)
      for curpoint,nextpoint in zip(a, b):
        yield curpoint, nextpoint

    newpaths = []
    from beziers.path import BezierPath
    for p in paths:
      newpath = []
      for scaledstart,scaledend in pairwise(p):
        key = (Point(*scaledstart), Point(*scaledend))
        if key in reconstructionLUT and not flat:
          orig = reconstructionLUT[key]
          if len(newpath) == 0 or newpath[-1] != orig:
            newpath.append(orig)
        else:
          newpath.append(Line(key[0]/precision, key[1]/precision))
      outpaths.append(BezierPath.fromSegments(newpath))
    return outpaths

  def union(self,other, flat=False):
    """Returns a list of Bezier paths representing the union of the two input paths."""
    return self.clip(other, pyclipper.CT_UNION, flat)

  def intersection(self,other, flat=False):
    """Returns a list of Bezier paths representing the intersection of the two input paths."""
    return self.clip(other, pyclipper.CT_INTERSECTION, flat)

  def difference(self,other, flat=False):
    """Returns a list of Bezier paths representing the first input path subtracted from the second."""
    return self.clip(other, pyclipper.CT_DIFFERENCE, flat)



from beziers.path.representations.Nodelist import Node

class GSPathRepresentation(object):
  def __init__(self, path, gspath = None):
    self.nodes = []
    from beziers.path import BezierPath
    assert isinstance(path, BezierPath)
    self.path = path
    if (gspath):
      self.nodes = gspath.nodes
      path.closed = gspath.closed

  def toNodelist(self):
    return list(map( lambda n: Node(n.position.x,n.position.y,n.type), self.nodes))



from beziers.line import Line
from beziers.cubicbezier import CubicBezier
from beziers.path import BezierPath
from beziers.path.representations.Nodelist import NodelistRepresentation, Node

class FontParts:
  @classmethod
  def fromFontpartsGlyph(klass, glyph):
    """Returns an *array of BezierPaths* from a FontParts glyph object."""
    paths = []
    if hasattr(glyph, "contours"):
      contouriterator = glyph.contours
    else:
      contouriterator = glyph
    for c in contouriterator:
      path = BezierPath()
      path.closed = False
      nodeList = []
      if hasattr(c, "points"):
        pointiterator = c.points
      else:
        pointiterator = c
      for p in pointiterator:
        if hasattr(p, "segmentType"):
          t = p.segmentType
        else:
          t = p.type
        nodeList.append(Node(p.x,p.y,t))
      path.activeRepresentation = NodelistRepresentation(path, nodeList)
      if nodeList[0].point == nodeList[-1].point:
        path.closed = True
      paths.append(path)
    return paths

  @classmethod
  def drawToFontpartsGlyph(klass,glyph,path):
    pen = glyph.getPen()
    segs = path.asSegments()
    pen.moveTo((segs[0][0].x,segs[0][0].y))
    for seg in segs:
      if isinstance(seg, Line):
        pen.lineTo((seg[1].x,seg[1].y))
      elif isinstance(seg, CubicBezier):
        pen.curveTo(
          (seg[1].x,seg[1].y),
          (seg[2].x,seg[2].y),
          (seg[3].x,seg[3].y))
    if path.closed:
      pen.closePath()



from beziers.segment import Segment
from beziers.line import Line
from beziers.point import Point
from beziers.utils import quadraticRoots, isclose
from beziers.utils.arclengthmixin import ArcLengthMixin

my_epsilon = 2e-7

class QuadraticBezier(ArcLengthMixin,Segment):
  def __init__(self, start, c1,end):
    self.points = [start,c1,end]
    self._range = [0,1]

  def __repr__(self):
    return "B<%s-%s-%s>" % (self[0],self[1],self[2])

  @classmethod
  def fromRepr(klass,text):
    import re
    p = re.compile("^B<(<.*?>)-(<.*?>)-(<.*?>)>$")
    m = p.match(text)
    points = [ Point.fromRepr(m.group(t)) for t in range(1,4) ]
    return klass(*points)

  def pointAtTime(self,t):
    """Returns the point at time t (0->1) along the curve."""
    x = (1 - t) * (1 - t) * self[0].x + 2 * (1 - t) * t * self[1].x + t * t * self[2].x;
    y = (1 - t) * (1 - t) * self[0].y + 2 * (1 - t) * t * self[1].y + t * t * self[2].y;
    return Point(x,y)

  def tOfPoint(self,p):
    """Returns the time t (0->1) of a point on the curve."""
    xroots = quadraticRoots(self[0].x - 2*self[1].x + self[2].x, 2 * (self[1].x-self[0].x), self[0].x - p.x)
    yroots = quadraticRoots(self[0].y - 2*self[1].y + self[2].y, 2 * (self[1].y-self[0].y), self[0].y - p.y)
    if not len(xroots) or not len(yroots):
      return -1
    for x in xroots:
      for y in yroots:
        if -my_epsilon < x - y < my_epsilon:
          return x
    return -1

  def splitAtTime(self,t):
    """Returns two segments, dividing the given segment at a point t (0->1) along the curve."""
    p4 = self[0].lerp(self[1],t)
    p5 = self[1].lerp(self[2],t)
    p7 = p4.lerp(p5,t)
    return (QuadraticBezier(self[0],p4,p7), QuadraticBezier(p7,p5,self[2]))

  def derivative(self):
    """Returns a `Line` representing the derivative of this curve."""
    return Line(
      (self[1]-self[0])*2,
      (self[2]-self[1])*2
    )

  def flatten(self, degree=8):
    samples = self.sample(self.length/degree)
    ss = []
    for i in range(1,len(samples)):
      l = Line(samples[i-1], samples[i])
      l._orig = self
      ss.append(l)
    return ss

  def _findRoots(self,dimension):
    if dimension == "x":
      return quadraticRoots(self[0].x - 2*self[1].x + self[2].x, 2 * (self[1].x-self[0].x), self[0].x)
    elif dimension == "y":
      return quadraticRoots(self[0].y - 2*self[1].y + self[2].y, 2 * (self[1].y-self[0].y), self[0].y)
    else:
      raise "Meh"

  def _findDRoots(self):
    d1 = (self[0].x-2*self[1].x+self[2].x)
    d2 = (self[0].y-2*self[1].y+self[2].y)
    roots = []
    if d1 != 0:
      r1 = (self[0].x-self[1].x)/d1
      roots.append(r1)
    if d2 != 0:
      r2 = (self[0].y-self[1].y)/d2
      roots.append(r2)
    return [ r for r in roots if r >= 0.01 and r <= 0.99 ]

  def findExtremes(self):
    """Returns a list of time `t` values for extremes of the curve."""
    return self._findDRoots()

  @property
  def area(self):
    """Returns the signed area between the curve and the y-axis"""
    return (2*(self[1].x*self[0].y - self[0].x*self[1].y - self[1].x*self[2].y + self[2].x*self[1].y) +
            3*(self[2].x*self[2].y - self[0].x*self[0].y) +
               self[2].x*self[0].y - self[0].x*self[2].y) / 6.

  def toCubicBezier(self):
    """Converts the quadratic bezier to a CubicBezier"""
    from beziers.cubicbezier import CubicBezier
    return CubicBezier(
      self[0], self[0]*(1/3.) + self[1]*(2/3.), self[1]*(2/3.) + self[2]*(1/3.), self[2]
    )



from beziers.point import Point
from beziers.affinetransformation import AffineTransformation
from beziers.utils.samplemixin import SampleMixin
from beziers.utils.intersectionsmixin import IntersectionsMixin
from beziers.boundingbox import BoundingBox

class Segment(IntersectionsMixin,SampleMixin,object):

  """A segment is part of a path. Although this package is called
  `beziers.py`, it's really for font people, and paths in the font
  world are made up of cubic Bezier curves, lines and (if you're
  dealing with TrueType) quadratic Bezier curves. Each of these
  things is represented as an object derived from the Segment base
  class. So, when you inspect the path in the segment representation,
  you will get a list of CubicBezier, Line and QuadraticBezier objects,
  all of which derive from Segment.

  Because of this, a Segment can have two, three or four elements:
  lines have two end points; quadratic Beziers have a start, a control
  point and an end point; cubic have a start, two control points and
  an end point.

  You can pretend that a Segment object is an array and index it like
  one::

      q = CubicBezier(
        Point(122,102), Point(35,200), Point(228,145), Point(190,46)
      )

      start, cp1, cp2, end = q[0],q[1],q[2],q[3]

  You can also access the start and end points like so::

      start = q.start
      end = q.end

  """

  def __getitem__(self, item):
    return self.points[item]
  def __setitem__(self, key, item):
      self.points[key] = item
  def __len__(self):
    return len(self.points)
  def __eq__(self,other):
    if self.order != other.order: return False
    for p in range(0,self.order):
      if self[p] != other[p]: return False
    return True
  def __hash__(self):
    return hash(tuple(self.points))
  def __ne__(self,other):
    return not self.__eq__(other)

  def clone(self):
    """Returns a new Segment which is a copy of this segment."""
    klass = self.__class__
    return klass(*[ p.clone() for p in self.points ])

  def round(self):
    """Rounds the points of segment to integer coordinates."""
    self.points = [ p.rounded() for p in self.points ]

  @property
  def order(self):
    return len(self.points)

  @property
  def start(self):
    """Returns a Point object representing the start of this segment."""
    return self.points[0]

  @property
  def end(self):
    """Returns a Point object representing the end of this segment."""
    return self.points[-1]

  @property
  def startAngle(self):
    return (self.points[1]-self.points[0]).angle

  @property
  def endAngle(self):
    return (self.points[-1]-self.points[-2]).angle

  def tangentAtTime(self,t):
    """Returns a `Point` representing the unit vector of tangent at time `t`."""
    return self.derivative().pointAtTime(t).toUnitVector()

  def normalAtTime(self,t):
    """Returns a `Point` representing the normal (rotated tangent) at time `t`."""
    tan = self.tangentAtTime(t)
    return Point(-tan.y,tan.x)

  def translated(self,vector):
    """Returns a *new Segment object* representing the translation of
    this segment by the given vector. i.e.::

      >>> l = Line(Point(0,0), Point(10,10))
      >>> l.translated(Point(5,5))
      L<<5.0,5.0>--<15.0,15.0>>
      >>> l
      L<<0.0,0.0>--<10.0,10.0>>

    """

    klass = self.__class__
    return klass(*[ p+vector for p in self.points ])

  def rotated(self,around, by):
    """Returns a *new Segment object* representing the rotation of
    this segment around the given point and by the given angle. i.e.::

      >>> l = Line(Point(0,0), Point(10,10))
      >>> l.rotated(Point(5,5), math.pi/2)
      L<<10.0,-8.881784197e-16>--<-8.881784197e-16,10.0>>

    """

    klass = self.__class__
    pNew = [ p.clone() for p in self.points]
    for p in pNew: p.rotate(around,by)
    return klass(*pNew)

  def scaled(self,bx):
    """Returns a *new Segment object* representing the scaling of
    this segment by the given magnification. i.e.::

      >>> l = Line(Point(0,0), Point(10,10))
      >>> l.scaled(2)
      L<<0,0>--<20,20>>

    """

    klass = self.__class__
    pNew = [ p * bx for p in self.points]
    return klass(*pNew)

  def transformed(self, transformation):
    """Returns a *new Segment object* transformed by the given AffineTransformation matrix."""
    klass = self.__class__
    pNew = [ p.clone() for p in self.points]
    for p in pNew: p.transform(transformation)
    return klass(*pNew)

  def alignmentTransformation(self):
    m = AffineTransformation.translation(self.start * -1)
    m.rotate((self.end.transformed(m)).angle * -1)
    return m

  def aligned(self):
    """Returns a *new Segment object* aligned to the origin. i.e.
    with the first point translated to the origin (0,0) and the
    last point with y=0. Obviously, for a `Line` this is a bit pointless,
    but it's quite handy for higher-order curves."""
    return self.transformed(self.alignmentTransformation())

  def lengthAtTime(self, t):
    """Returns the length of the subset of the path from the start
    up to the point t (0->1), where 1 is the end of the whole curve."""
    s1,_ = self.splitAtTime(t)
    return s1.length

  def reversed(self):
    """Returns a new segment with the points reversed."""
    klass = self.__class__
    return klass(*list(reversed(self.points)))

  def bounds(self):
    """Returns a BoundingBox object for this segment."""
    bounds = BoundingBox()
    ex = self.findExtremes()
    ex.append(0)
    ex.append(1)
    for t in ex:
      bounds.extend(self.pointAtTime(t))
    return bounds

  @property
  def hasLoop(self):
    """Returns True if the segment has a loop. (Only possible for cubics.)"""
    return False
  



import unittest
from beziers.quadraticbezier import QuadraticBezier
from beziers.cubicbezier import CubicBezier
from beziers.point import Point

class ArcLengthMethods(unittest.TestCase):
  def test_quadraticLength(self):
    b1 = QuadraticBezier(
      Point(150,40),
      Point(80,30),
      Point(105,150)
    )
    self.assertAlmostEqual(b1.length, 144.79561403359523)

    b2 = QuadraticBezier(
      Point(707,596),
      Point(645,596),
      Point(592,623)
    )
    self.assertAlmostEqual(b2.length, 119.25113694489232)

  def test_cubicLength(self):
    b1 = CubicBezier(
        Point(100,25),
        Point(10,90),
        Point(110,100),
        Point(132,192)
      )
    self.assertAlmostEqual(b1.length, 202.20118972656385)



import unittest
from beziers.utils.curvefitter import CurveFit
from beziers.path import BezierPath
from beziers.point import Point
from beziers.path.representations.Nodelist import NodelistRepresentation, Node

class CornerMethods(unittest.TestCase):
  def test_corners(self):
    nl = [
      Node(302.0,492.0,"line"),
      Node(176.0,432.0,"line"),
      Node(-51.0,325.0,"offcurve"),
      Node(-74.0,484.0,"offcurve"),
      Node(73.0,570.0,"curve"),
      Node(85.0,764.0,"offcurve"),
      Node(290.0,748.0,"offcurve"),
      Node(418.0,688.0,"curve"),
    ]
    path = BezierPath.fromNodelist(nl)
    path.closed = False
    for seg1, seg2 in path.segpairs():
      print(seg1.endAngle * 57.2958, seg2.startAngle * 57.2958)



from shapely.ops import cascaded_union, polygonize
from scipy.spatial import Delaunay
import shapely.geometry as geometry
import numpy as np
import math

def alpha_shape(points, alpha):
    """
    Compute the alpha shape (concave hull) of a set
    of points.
    @param points: Iterable container of points.
    @param alpha: alpha value to influence the
        gooeyness of the border. Smaller numbers
        don't fall inward as much as larger numbers.
        Too large, and you lose everything!
    """
    if len(points) < 4:
        # When you have a triangle, there is no sense
        # in computing an alpha shape.
        return geometry.MultiPoint(list(points)).convex_hull

    coords = np.array([point.coords[0] for point in points])
    tri = Delaunay(coords)
    triangles = coords[tri.vertices]
    a = ((triangles[:,0,0] - triangles[:,1,0]) ** 2 + (triangles[:,0,1] - triangles[:,1,1]) ** 2) ** 0.5
    b = ((triangles[:,1,0] - triangles[:,2,0]) ** 2 + (triangles[:,1,1] - triangles[:,2,1]) ** 2) ** 0.5
    c = ((triangles[:,2,0] - triangles[:,0,0]) ** 2 + (triangles[:,2,1] - triangles[:,0,1]) ** 2) ** 0.5
    s = ( a + b + c ) / 2.0
    areas = (s*(s-a)*(s-b)*(s-c)) ** 0.5
    circums = a * b * c / (4.0 * areas)
    filtered = triangles[circums < (1.0 / alpha)]
    edge1 = filtered[:,(0,1)]
    edge2 = filtered[:,(1,2)]
    edge3 = filtered[:,(2,0)]
    edge_points = np.unique(np.concatenate((edge1,edge2,edge3)), axis = 0).tolist()
    m = geometry.MultiLineString(edge_points)
    triangles = list(polygonize(m))
    return cascaded_union(triangles), edge_points



from beziers.path.representations.Segment import SegmentRepresentation
from beziers.path.representations.Nodelist import NodelistRepresentation, Node
from beziers.point import Point
from beziers.boundingbox import BoundingBox
from beziers.utils.samplemixin import SampleMixin
from beziers.utils.booleanoperationsmixin import BooleanOperationsMixin
from beziers.segment import Segment
from beziers.line import Line
from beziers.cubicbezier import CubicBezier

import math

if not hasattr(math, "isclose"):
  def isclose(a, b, rel_tol=1e-9, abs_tol=0.0):
    return abs(a-b) <= max(rel_tol * max(abs(a), abs(b)), abs_tol)
  math.isclose = isclose

class BezierPath(BooleanOperationsMixin,SampleMixin,object):
  """`BezierPath` represents a collection of `Segment` objects - the
  curves and lines that make up a path.

  One of the really fiddly things about manipulating Bezier paths in
  a computer is that there are various ways to represent them.
  Different applications prefer different representations. For instance,
  when you're drawing a path on a canvas, you often want a list of nodes
  like so::

    { "x":255.0, "y":20.0, "type":"curve"},
    { "x":385.0, "y":20.0, "type":"offcurve"},
    { "x":526.0, "y":79.0, "type":"offcurve"},
    { "x":566.0, "y":135.0, "type":"curve"},
    { "x":585.0, "y":162.0, "type":"offcurve"},
    { "x":566.0, "y":260.0, "type":"offcurve"},
    { "x":484.0, "y":281.0, "type":"curve"},
    ...

  But when you're doing Clever Bezier Mathematics, you generally want
  a list of segments instead::

    [ (255.0,20.0), (385.0,20.0), (526.0,79.0), (566.0,135.0)],
    [ (566.0,135.0), (585.0,162.0), (566.0,260.0), (484.0,281.0)],

  The Beziers module is designed to allow you to move fluidly between these
  different representations depending on what you're wanting to do.

  """

  def __init__(self):
    self.activeRepresentation = None
    self.closed = True


  @classmethod
  def fromPoints(self, points, error = 50.0, cornerTolerance = 20.0, maxSegments = 20):
    """Fit a poly-bezier curve to the points given. This operation should be familiar
    from 'pencil' tools in a vector drawing application: the application samples points
    where your mouse pointer has been dragged, and then turns the sketch into a Bezier
    path. The goodness of fit can be controlled by tuning the `error` parameter. Corner
    detection can be controlled with `cornerTolerance`.

    Here are some points fit with `error=100.0`:

..  figure:: curvefit1.png
    :scale: 75 %
    :alt: curvefit1


    And with `error=10.0`:

..  figure:: curvefit2.png
    :scale: 75 %
    :alt: curvefit1

    """
    from beziers.utils.curvefitter import CurveFit
    segs = CurveFit.fitCurve(points, error, cornerTolerance, maxSegments)
    path = BezierPath()
    path.closed = False
    path.activeRepresentation = SegmentRepresentation(path,segs)
    return path

  @classmethod
  def fromSegments(klass, array):
    """Construct a path from an array of Segment objects."""
    self = klass()
    for a in array: assert(isinstance(a, Segment))
    self.activeRepresentation = SegmentRepresentation(self,array)
    return self

  @classmethod
  def fromNodelist(klass, array, closed=True):
    """Construct a path from an array of Node objects."""
    self = klass()
    for a in array: assert(isinstance(a, Node))
    self.closed = closed
    self.activeRepresentation = NodelistRepresentation(self,array)
    self.asSegments() # Resolves a few problems
    return self

  @classmethod
  def fromGlyphsLayer(klass, layer):
    """Returns an *array of BezierPaths* from a Glyphs GSLayer object."""
    from beziers.path.representations.GSPath import GSPathRepresentation

    bezpaths = []
    for p in layer.paths:
      path = BezierPath()
      path.activeRepresentation = GSPathRepresentation(path,p)
      bezpaths.append(path)
    return bezpaths

  @classmethod
  def fromDrawable(klass, layer, *penArgs, **penKwargs):
    """Returns an *array of BezierPaths* from any object conforming to the pen protocol."""
    from beziers.utils.pens import BezierPathCreatingPen
    pen = BezierPathCreatingPen(*penArgs, **penKwargs)
    layer.draw(pen)
    return pen.paths

  @classmethod
  def fromFonttoolsGlyph(klass,font,glyphname):
    """Returns an *array of BezierPaths* from a FontTools font object and glyph name."""
    glyphset = font.getGlyphSet()
    from beziers.utils.pens import BezierPathCreatingPen
    pen = BezierPathCreatingPen(glyphset)
    _glyph = font.getGlyphSet()[glyphname]._glyph
    if "glyf" in font:
      _glyph.draw(pen, font["glyf"])
    else:
      _glyph.draw(pen)
    return pen.paths

  def asSegments(self):
    """Return the path as an array of segments (either Line, CubicBezier,
    or, if you are exceptionally unlucky, QuadraticBezier objects)."""
    if not isinstance(self.activeRepresentation, SegmentRepresentation):
      nl = self.activeRepresentation.toNodelist()
      assert isinstance(nl, list)
      self.activeRepresentation = SegmentRepresentation.fromNodelist(self,nl)
    return self.activeRepresentation.data()

  def asNodelist(self):
    """Return the path as an array of Node objects."""
    if not isinstance(self.activeRepresentation, NodelistRepresentation):
      nl = self.activeRepresentation.toNodelist()
      assert isinstance(nl, list)
      self.activeRepresentation = NodelistRepresentation(self,nl)
    return self.activeRepresentation.data()

  def asSVGPath(self):
    """Return the path as a string suitable for a SVG <path d="..."? element."""
    segs = self.asSegments()
    pathParts = ["M %f %f" % (segs[0][0].x, segs[0][0].y)]

    operators = "xxLQC"
    for s in segs:
      op = operators[len(s)] + " "
      for pt in s[1:]:
        op = op + "%f %f " % (pt.x, pt.y)
      pathParts.append(op)
    if self.closed:
      pathParts.append("Z")

    return " ".join(pathParts)

  def asMatplot(self):
    from matplotlib.path import Path
    nl = self.asNodelist()
    verts = [(nl[0].x,nl[0].y)]
    codes = [Path.MOVETO]

    for i in range(1,len(nl)):
      n = nl[i]
      verts.append((n.x,n.y))
      if n.type == "offcurve":
        if nl[i+1].type == "offcurve" or nl[i-1].type == "offcurve":
          codes.append(Path.CURVE4)
        else:
          codes.append(Path.CURVE3)
      elif n.type == "curve":
        if nl[i-1].type == "offcurve" and i >2 and nl[i-2].type == "offcurve":
          codes.append(Path.CURVE4)
        else:
          codes.append(Path.CURVE3)
      elif n.type == "line":
        codes.append(Path.LINETO)
      else:
        raise "Unknown node type"
    if self.closed:
      verts.append((nl[0].x,nl[0].y))
      codes.append(Path.CLOSEPOLY)
    return Path(verts, codes)

  def plot(self,ax, **kwargs):
    """Plot the path on a Matplot subplot which you supply

    ::

          import matplotlib.pyplot as plt
          fig, ax = plt.subplots()
          path.plot(ax)

    """
    import matplotlib.pyplot as plt
    from matplotlib.lines import Line2D
    from matplotlib.path import Path
    import matplotlib.patches as patches
    path = self.asMatplot()
    if not "lw" in kwargs:
      kwargs["lw"] = 2
    if not "fill" in kwargs:
      kwargs["fill"] = False
    drawNodes = (not("drawNodes" in kwargs) or kwargs["drawNodes"] != False)
    if "drawNodes" in kwargs:
      kwargs.pop("drawNodes")
    patch = patches.PathPatch(path, **kwargs)
    ax.add_patch(patch)
    left, right = ax.get_xlim()
    top, bottom = ax.get_ylim()
    bounds = self.bounds()
    bounds.addMargin(50)
    if not (left == 0.0 and right == 1.0 and top == 0.0 and bottom == 1.0):
      bounds.extend(Point(left,top))
      bounds.extend(Point(right,bottom))
    ax.set_xlim(bounds.left,bounds.right)
    ax.set_ylim(bounds.bottom,bounds.top)
    if drawNodes:
      nl = self.asNodelist()
      for i in range(0,len(nl)):
        n = nl[i]
        if n.type =="offcurve":
          circle = plt.Circle((n.x, n.y), 2, fill=True,color="black",alpha=0.5)
          ax.add_artist(circle)
          if i+1 < len(nl) and nl[i+1].type != "offcurve":
            l = Line2D([n.x, nl[i+1].x], [n.y, nl[i+1].y], linewidth=2,color="black",alpha=0.3)
            ax.add_artist(l)
          if i-0 >= 0 and nl[i-1].type != "offcurve":
            l = Line2D([n.x, nl[i-1].x], [n.y, nl[i-1].y], linewidth=2,color="black",alpha=0.3)
            ax.add_artist(l)
        else:
          circle = plt.Circle((n.x, n.y), 3,color="black",alpha=0.3)
          ax.add_artist(circle)

  def clone(self):
    """Return a new path which is an exact copy of this one"""
    p = BezierPath.fromSegments(self.asSegments())
    p.closed = self.closed
    return p

  def round(self):
    """Rounds the points of this path to integer coordinates."""
    segs = self.asSegments()
    for s in segs: s.round()
    self.activeRepresentation = SegmentRepresentation(self,segs)

  def bounds(self):
    """Determine the bounding box of the path, returned as a
    `BoundingBox` object."""
    bbox = BoundingBox()
    for seg in self.asSegments():
      bbox.extend(seg)
    return bbox

  def splitAtPoints(self,splitlist):
    def mapx(v,ds): return (v-ds)/(1-ds)
    segs = self.asSegments()
    newsegs = []
    # Cluster splitlist by seg
    newsplitlist = {}
    for (seg,t) in splitlist:
      if not seg in newsplitlist: newsplitlist[seg] = []
      newsplitlist[seg].append(t)
    for k in newsplitlist:
      newsplitlist[k] = sorted(newsplitlist[k])
    # Now walk the path
    for seg in segs:
      if seg in newsplitlist:
        tList = newsplitlist[seg]
        while len(tList) > 0:
          t = tList.pop(0)
          if t < 1e-8: continue
          seg1,seg2 = seg.splitAtTime(t)
          newsegs.append(seg1)
          seg = seg2
          for i in range(0,len(tList)): tList[i] = mapx(tList[i],t)
      newsegs.append(seg)
    self.activeRepresentation = SegmentRepresentation(self,newsegs)


  def addExtremes(self):
    """Add extreme points to the path."""
    def mapx(v,ds): return (v-ds)/(1-ds)
    segs = self.asSegments()
    splitlist = []
    for seg in segs:
      for t in seg.findExtremes():
        splitlist.append((seg,t))
    self.splitAtPoints(splitlist)
    return self

  @property
  def length(self):
    """Returns the length of the whole path."""
    segs = self.asSegments()
    length = 0
    for s in segs: length += s.length
    return length

  def pointAtTime(self,t):
    """Returns the point at time t (0->1) along the curve, where 1 is the end of the whole curve."""
    segs = self.asSegments()
    if t == 1.0:
      return segs[-1].pointAtTime(1)
    t *= len(segs)
    seg = segs[int(math.floor(t))]
    return seg.pointAtTime(t-math.floor(t))

  def lengthAtTime(self,t):
    """Returns the length of the subset of the path from the start
    up to the point t (0->1), where 1 is the end of the whole curve."""
    segs = self.asSegments()
    t *= len(segs)
    length = 0
    for s in segs[:int(math.floor(t))]: length += s.length
    seg = segs[int(math.floor(t))]
    s1,s2 = seg.splitAtTime(t-math.floor(t))
    length += s1.length
    return length

  def offset(self, vector, rotateVector = True):
    """Returns a new BezierPath which approximates offsetting the
    current Bezier path by the given vector. Note that the vector
    will be rotated around the normal of the curve so that the
    offsetting always happens on the same 'side' of the curve:

..  figure:: offset1.png
    :scale: 75 %
    :alt: offset1

    If you don't want that and you want 'straight' offsetting instead
    (which may intersect with the original curve), pass
    `rotateVector=False`:

..  figure:: offset2.png
    :scale: 75 %
    :alt: offset1

    """
    # Method 1 - curve fit
    newsegs = []
    points = []
    def finishPoints(newsegs, points):
      if len(points) > 0:
        bp = BezierPath.fromPoints(points, error=0.1, cornerTolerance= 1)
        newsegs.extend(bp.asSegments())
      while len(points)>0:points.pop()

    for seg in self.asSegments():
      if isinstance(seg,Line):
        finishPoints(newsegs,points)
        newsegs.append(seg.translated(vector))
      else:
        t = 0.0
        while t <1.0:
          if rotateVector:
            points.append( seg.pointAtTime(t) + vector.rotated(Point(0,0), seg.normalAtTime(t).angle))
          else:
            points.append( seg.pointAtTime(t) + vector)
          step = max(abs(seg.curvatureAtTime(t)),0.1)
          t = t + min(seg.length / step,0.1)
    finishPoints(newsegs,points)
    newpath = BezierPath()
    newpath.activeRepresentation = SegmentRepresentation(newpath, newsegs)
    return newpath

  def append(self, other, joinType="line"):
    """Append another path to this one. If the end point of the first
    path is not the same as the start point of the other path, a line
    will be drawn between them."""
    segs1 = self.asSegments()
    segs2 = other.asSegments()
    if len(segs1) < 1:
      self.activeRepresentation = SegmentRepresentation(self, segs2)
      return
    if len(segs2) < 1:
      self.activeRepresentation = SegmentRepresentation(self, segs1)
      return

    # Which way around should they go?
    dist1 = segs1[-1].end.distanceFrom(segs2[0].start)
    dist2 = segs1[-1].end.distanceFrom(segs2[-1].end)
    if dist2 > 2 * dist1:
      segs2 = list(reversed([ x.reversed() for x in segs2]))

    # Add a line between if they don't match up
    if segs1[-1].end != segs2[0].start:
      segs1.append(Line(segs1[-1].end,segs2[0].start))

    # XXX Check for discontinuities and harmonize if needed

    segs1.extend(segs2)
    self.activeRepresentation = SegmentRepresentation(self, segs1)
    return self

  def reverse(self):
    """Reverse this path (mutates path)."""
    seg2 = [ x.reversed() for x in self.asSegments()]
    self.activeRepresentation = SegmentRepresentation(self, list(reversed(seg2)))
    return self

  def translate(self, vector):
    """Translates the path by a given vector."""
    seg2 = [ x.translated(vector) for x in self.asSegments()]
    self.activeRepresentation = SegmentRepresentation(self, seg2)
    return self

  def rotate(self, about, angle):
    """Rotate the path by a given vector."""
    seg2 = [ x.rotated(about, angle) for x in self.asSegments()]
    self.activeRepresentation = SegmentRepresentation(self, seg2)
    return self

  def scale(self, by):
    """Scales the path by a given magnitude."""
    seg2 = [ x.scaled(by) for x in self.asSegments()]
    self.activeRepresentation = SegmentRepresentation(self, seg2)
    return self

  def balance(self):
    """Performs Tunni balancing on the path."""
    segs = self.asSegments()
    for x in segs:
      if isinstance(x, CubicBezier): x.balance()
    self.activeRepresentation = SegmentRepresentation(self, segs)
    return self

  def findDiscontinuities(self):
    """Not implemented yet"""

  def harmonize(self):
    """Not implemented yet"""

  def roundCorners(self):
    """Not implemented yet"""

  def dash(self, lineLength = 50, gapLength = None):
    """Returns a list of BezierPath objects created by chopping
    this path into a dashed line::

      paths = path.dash(lineLength = 20, gapLength = 50)

..  figure:: dash.png
    :scale: 75 %
    :alt: path.dash(lineLength = 20, gapLength = 50)

"""
    if not gapLength:
      gapLength = lineLength
    granularity = self.length
    newpaths = []
    points = []
    for t in self.regularSampleTValue(granularity):
      lenSoFar = self.lengthAtTime(t) # Super inefficient. But simple!
      lenSoFar = lenSoFar % (lineLength + gapLength)
      if lenSoFar >= lineLength and len(points) > 0:
        # When all you have is a hammer...
        bp = BezierPath.fromPoints(points, error=1, cornerTolerance= 1)
        points = []
        if len(bp.asSegments()) > 0: newpaths.append(bp)
      elif lenSoFar <= lineLength:
        points.append(self.pointAtTime(t))
    return newpaths

  def segpairs(self):
    segs = self.asSegments()
    for i in range(0,len(segs)-1):
      yield (segs[i],segs[i+1])

  def harmonize(self, seg1, seg2):
    if seg1.end.x != seg2.start.x or seg1.end.y != seg2.start.y: return
    a1, a2 = seg1[1], seg1[2]
    b1, b2 = seg2[1], seg2[2]
    intersections = Line(a1,a2).intersections(Line(b1,b2),limited=False)
    if not intersections[0]: return
    p0 = a1.distanceFrom(a2) / a2.distanceFrom(intersections[0].point)
    p1 = b1.distanceFrom(intersections[0].point) / b1.distanceFrom(b2)
    r = math.sqrt(p0 * p1)
    t = r / (r+1)
    newA3 = a2.lerp(b1, t)
    fixup = seg2.start - newA3
    seg1[2] += fixup
    seg2[1] += fixup

  def flatten(self,degree=8):
    segs = []
    for s in self.asSegments():
      segs.extend(s.flatten(degree))
    return BezierPath.fromSegments(segs)

  def windingNumberOfPoint(self,pt):
    bounds = self.bounds()
    bounds.addMargin(10)
    ray1 = Line(Point(bounds.left,pt.y),pt)
    ray2 = Line(Point(bounds.right,pt.y),pt)
    leftIntersections = {}
    rightIntersections = {}
    leftWinding = 0
    rightWinding = 0
    for s in self.asSegments():
      for i in s.intersections(ray1):
        # print("Found left intersection with %s: %s" % (ray1, i.point))
        leftIntersections[i.point] = i

      for i in s.intersections(ray2):
        rightIntersections[i.point] = i

    for i in leftIntersections.values():
      # XXX tangents here are all positive? Really?
      # print(i.seg1, i.t1, i.point)
      tangent = s.tangentAtTime(i.t1)
      # print("Tangent at left intersection %s is %f" % (i.point,tangent.y))
      leftWinding += int(math.copysign(1,tangent.y))

    for i in rightIntersections.values():
      tangent = s.tangentAtTime(i.t1)
      # print("Tangent at right intersection %s is %f" % (i.point,tangent.y))
      rightWinding += int(math.copysign(1,tangent.y))

    # print("Left winding: %i right winding: %i " % (leftWinding,rightWinding))
    return max(abs(leftWinding),abs(rightWinding))

  def pointIsInside(self,pt):
    """Returns true if the given point lies on the "inside" of the path,
    assuming an 'even-odd' winding rule where self-intersections are considered
    outside."""
    li = self.windingNumberOfPoint(pt)
    return li % 2 == 1

  @property
  def signed_area(self):
    """Approximates the area under a closed path by flattening and treating as a polygon.
    Returns the signed area; positive means the path is counter-clockwise,
    negative means it is clockwise."""
    flat = self.flatten()
    area = 0
    for s in flat.asSegments():
      area = area + (s.start.x * s.end.y) - (s.start.y * s.end.x)
    area = area / 2.0
    return area

  @property
  def area(self):
    """Approximates the area under a closed path by flattening and treating as a
    polygon. Returns the unsigned area. Use :py:meth:`signed_area` if you want
    the signed area."""
    return abs(self.signed_area)

  @property
  def direction(self):
    """Returns the direction of the path: -1 for clockwise and 1 for counterclockwise."""
    return math.copysign(1, self.signed_area)

  @property
  def centroid(self):
    if not self.closed: return None
    return self.bounds().centroid # Really?

  def drawWithBrush(self, other):
    """Assuming that `other` is a closed Bezier path representing a pen or
    brush of a certain shape and that `self` is an open path, this method
    traces the brush along the path, returning an array of Bezier paths.

    `other` may also be a function which, given a time `t` (0-1), returns a closed
    path representing the shape of the brush at the given time.

    This requires the `shapely` library to be installed.
    """
    from shapely.geometry import Polygon
    from shapely.ops import unary_union
    polys = []
    samples = self.sample(self.length/2)
    def constantBrush(t):
      return other

    brush = other
    if not callable(brush):
      brush = constantBrush

    c = brush(0).centroid

    from itertools import tee
    def pairwise(iterable):
        "s -> (s0,s1), (s1,s2), (s2, s3), ..."
        a, b = tee(iterable)
        next(b, None)
        return zip(a, b)

    t = 0
    for n in samples:
      brushHere = brush(t).clone().flatten()
      brushHere.translate(n-brushHere.centroid)
      polys.append( Polygon([ (x[0].x, x[0].y) for x in brushHere.asSegments() ]) )
      t = t + 1.0/len(samples)
    concave_hull = unary_union(polys)
    ll = []
    for x,y in pairwise(concave_hull.exterior.coords):
      l = Line(Point(x[0],x[1]),Point(y[0],y[1]))
      ll.append(l)
    paths = [ BezierPath.fromSegments(ll) ]

    for interior in concave_hull.interiors:
      ll = []
      for x,y in pairwise(interior.coords):
        l = Line(Point(x[0],x[1]),Point(y[0],y[1]))
        ll.append(l)
      paths.append( BezierPath.fromSegments(ll) )

    return paths

  def quadraticsToCubics(self):
    """Converts all quadratic segments in the path to cubic Beziers."""
    segs = self.asSegments()
    for i,s in enumerate(segs):
      if len(s) == 3:
        segs[i] = s.toCubicBezier()
    return self

  def thicknessAtX(path, x):
    """Returns the thickness of the path at x-coordinate ``x``."""
    bounds = path.bounds()
    bounds.addMargin(10)
    ray = Line(Point(x - 0.1, bounds.bottom), Point(x + 0.1, bounds.top))
    intersections = []
    for seg in path.asSegments():
      intersections.extend(seg.intersections(ray))
    if len(intersections) < 2:
      return None
    intersections = list(sorted(intersections, key=lambda i: i.point.y))
    i1, i2 = intersections[0:2]
    inorm1 = i1.seg1.normalAtTime(i1.t1)
    ray1 = Line(i1.point + (inorm1 * 1000), i1.point + (inorm1 * -1000))
    iii = i2.seg1.intersections(ray1)
    if iii:
      ll1 = i1.point.distanceFrom(iii[0].point)
    else:
      # Simple, vertical version
      return abs(i1.point.y - i2.point.y)

    inorm2 = i2.seg1.normalAtTime(i2.t1)
    ray2 = Line(i2.point + (inorm2 * 1000), i2.point + (inorm2 * -1000))
    iii = i1.seg1.intersections(ray2)
    if iii:
      ll2 = i2.point.distanceFrom(iii[0].point)
      return (ll1 + ll2) * 0.5
    else:
      return ll1

    # midpoint = (i1.point + i2.point) / 2
    # # Find closest path to midpoint
    # # Find the tangent at that time
    # inorm2 = i2.seg1.normalAtTime(i2.t1)

  def distanceToPath(self, other, samples = 10):
    """Finds the distance to the other curve at its closest point,
    along with the t values for the closest point at each segment
    and the relevant segments.

    Returns: ``distance, t1, t2, seg1, seg2``."""
    from beziers.utils.curvedistance import curveDistance
    segs1 = self.asSegments()
    segs2 = other.asSegments()
    minDistance = None
    # Find closest segment pair.
    for s1 in segs1:
      samples1 = s1.sample(samples)
      for s2 in segs2:
        samples2 = s2.sample(samples)
        d = min([ p1.squareDistanceFrom(p2) for p1 in samples1 for p2 in samples2])
        if not minDistance or d < minDistance:
          minDistance = d
          closestPair = (s1,s2)
    c = curveDistance(closestPair[0], closestPair[1])
    return (c[0],c[1],c[2], closestPair[0], closestPair[1])

  def tidy(self, **kwargs):
    """Tidies a curve by adding extremes, and then running
    ``removeIrrelevantSegments`` and ``smooth``. ``relLength``,
    ``absLength``, ``maxCollectionSize``, ``lengthLimit`` and
    ``cornerTolerance`` parameters are passed to the relevant routine."""
    self.addExtremes()
    self.removeIrrelevantSegments(**{k:v for k,v in kwargs.items()
      if k in ["relLength", "absLength"] })
    self.smooth(**{k:v for k,v in kwargs.items()
      if k in ["maxCollectionSize", "lengthLimit", "cornerTolerance"] })

  def removeIrrelevantSegments(self, relLength=1/50000, absLength=0):
    """Removes small and collinear line segments. Collinear line
    segments are adjacent line segments which are heading in the same
    direction, and hence can be collapsed into a single segment.
    Small segments (those less than ``absLength`` units, or less than
    ``relLength`` as a fraction of the path's total length) are
    removed entirely."""
    segs = self.asSegments()
    newsegs = [segs[0]]
    smallLength = self.length * relLength
    for i in range(1,len(segs)):
      prev = newsegs[-1]
      this = segs[i]
      if this.length < smallLength or this.length < absLength:
        this[0] = prev[0]
        newsegs[-1] = this
        continue
      if len(prev) == 2 and len(this) == 2:
        if math.isclose(prev.tangentAtTime(0).angle, this.tangentAtTime(0).angle):
          this[0] = prev[0]
          newsegs[-1] = this
          continue
      newsegs.append(this)
    self.activeRepresentation = SegmentRepresentation(self, newsegs)
    return self

  def smooth(self, maxCollectionSize = 30, lengthLimit = 20, cornerTolerance = 10):
    """Smooths a curve, by collating lists of small (at most ``lengthLimit``
    units long) segments at most ``maxCollectionSize`` segments at a time,
    and running them through a curve fitting algorithm. The list collation
    also stops when one segment turns more than ``cornerTolerance`` degrees
    away from the previous one, so that corners are not smoothed."""
    smallLineLength = lengthLimit
    segs = self.asSegments()
    i = 0
    collection = []
    while i < len(segs):
        s = segs[i]
        if s.length < smallLineLength and len(collection) <= maxCollectionSize:
          collection.append(s)
        else:
          corner = False
          if len(collection) > 1:
            last = collection[-1]
            if abs(last.tangentAtTime(1).angle - s.tangentAtTime(0).angle) > math.radians(cornerTolerance):
              corner = True
          if len(collection) > maxCollectionSize or corner or i == len(segs)-2:
            points = [x.start for x in collection]
            bp = BezierPath.fromPoints(points)
            if len(bp.asSegments()) > 0:
              segs[i-len(collection):i] = bp.asSegments()
              i -= len(collection)
            collection = []
        i += 1
    if len(collection) > 0:
      points = [x.start for x in collection]
      bp = BezierPath.fromPoints(points)
      if len(bp.asSegments()) > 0:
        segs[i-(1+len(collection)):i-1] = bp.asSegments()

    self.activeRepresentation = SegmentRepresentation(self, segs)
    return self



from beziers.cubicbezier import CubicBezier
from beziers.quadraticbezier import QuadraticBezier
from beziers.line import Line
from beziers.point import Point
import math

"""
This implements (possibly badly) the algorithm in "Computing the
minimum distance between two Bézier curves", Chen et al.,
*Journal of Computational and Applied Mathematics* 229(2009) 294-301

I could not get the clever subdivision heuristic working (divide the
parameter space into two) but divided it into four instead. Using
their heuristic of dividing at i/2n instead of 1/2 does seem to give
better results, though. The checks for Property 1 and Property 2 never
seem to be fired either. So I'm probably doing something wrong.

The results look fine but there's probably more computations happening
than needed. I use memoization and an initial sanity check (whether
alpha is more than the current best alpha) to cut down on the
computations but I'm sure it could still be improved.
"""

def C(y, x):
    try:
        binom = math.factorial(x) // math.factorial(y) // math.factorial(x - y)
    except ValueError:
        binom = 0
    return binom


def A_r(r, P):
    n = len(P) - 1  # Order
    upsilon = min(r, n)
    theta = max(0, r - n)
    summand = 0
    for i in range(theta, upsilon + 1):
        summand += P[i].dot(P[r - i]) * C(i, n) * C(r - i, n) / C(r, 2 * n)
    return summand


B_k = A_r  # But with P -> Q, n -> m, r -> k


def C_rk(r, k, bez1, bez2):
    # These also have parallel factors
    def factor(P, r_or_k):
        n = len(P) - 1
        upsilon = min(r_or_k, n)
        theta = max(0, r_or_k - n)
        summand = Point(0, 0)
        for i in range(theta, upsilon + 1):
            summand += P[i] * C(i, n) * C(r_or_k - i, n) / C(r_or_k, 2 * n)
        return summand

    left = factor(bez1, r)
    right = factor(bez2, k)
    return left.dot(right)


def basis_function(n, i, u):
    res = C(i, n) * (1 - u) ** (n - i) * u ** i
    return res


class MinimumCurveDistanceFinder:
    def __init__(self, bez1, bez2):
        self.bez1 = bez1
        self.bez2 = bez2
        self.dCache = {}
        self.sCache = {}
        self.bestAlpha = None
        self.iterations = 0

    def D(self, r, k):
        if (r, k) not in self.dCache:
            self.dCache[(r, k)] = (
                A_r(r, self.bez1)
                + B_k(k, self.bez2)
                - 2 * C_rk(r, k, self.bez1, self.bez2)
            )
        return self.dCache[(r, k)]

    def S(self, u, v):  # u,v are times
        if (u, v) in self.sCache:
            return self.sCache[(u, v)]

        n = len(self.bez1) - 1
        m = len(self.bez2) - 1

        summand = 0
        for r in range(0, 2 * n + 1):
            for k in range(0, 2 * m + 1):
                summand += (
                    self.D(r, k)
                    * basis_function(2 * n, r, u)
                    * basis_function(2 * m, k, v)
                )
        self.sCache[(u, v)] = summand
        return summand

    def minDist(self, uinterval=(0, 1), vinterval=(0, 1), epsilon=0.001):
        n = len(self.bez1) - 1
        m = len(self.bez2) - 1
        self.iterations = self.iterations + 1
        umin, umax = uinterval
        vmin, vmax = vinterval
        umid = (umin + umax) / 2
        vmid = (vmin + vmax) / 2
        svalues = [
            [self.S(umin, vmin), umin, vmin],
            [self.S(umin, vmax), umin, vmax],
            [self.S(umax, vmin), umax, vmin],
            [self.S(umax, vmax), umax, vmax],
        ]
        alpha = min(svalues, key=lambda x: x[0])[0]
        if self.bestAlpha and alpha > self.bestAlpha:
            return [alpha, umid, vmid]
        self.bestAlpha = alpha

        if abs(umax - umin) <= epsilon or abs(vmax - vmin) <= epsilon:
            return [alpha, umid, vmid]

        # # Property 1: Check if any D(r>k) < alpha
        # # Also find out where the min value is for division
        isOutside = True
        minDRK = None
        minIJ = (None, None)
        for r in range(0, 2 * n):
            for k in range(0, 2 * m):
                drk = self.D(r, k)
                if drk < alpha:
                    isOutside = False
                if not minDRK or drk < minDRK:
                    minDrk = drk
                    minIJ = (r, k)
        if isOutside:
            return [alpha, umid, vmid]

        # This one never seems to work

        # # Property 2: Boundary check
        atBoundary0onBez1 = True
        atBoundary1onBez1 = True
        atBoundary0onBez2 = True
        atBoundary1onBez2 = True
        for i in range(0, 2 * n):
            for j in range(0, 2 * m):
                dij = self.D(i, j)
                dkj = self.D(0, j)
                if dij < dkj:
                    atBoundary0onBez1 = False
                dkj = self.D(2 * n, j)
                if dij < dkj:
                    atBoundary1onBez1 = False
                dkj = self.D(i, 0)
                if dij < dkj:
                    atBoundary0onBez2 = False
                dkj = self.D(i, 2 * n)
                if dij < dkj:
                    atBoundary1onBez2 = False
        if atBoundary0onBez1 and atBoundary0onBez2:
            return svalues[0]
        if atBoundary0onBez1 and atBoundary1onBez2:
            return svalues[1]
        if atBoundary1onBez1 and atBoundary0onBez2:
            return svalues[2]
        if atBoundary1onBez1 and atBoundary1onBez2:
            return svalues[3]

        newuMid = umin + (umax - umin) * (minIJ[0] / (2 * n))
        newvMid = vmin + (vmax - vmin) * (minIJ[1] / (2 * m))

        # Subdivision
        r1 = self.minDist(uinterval=(umin, newuMid), vinterval=(vmin, newvMid))
        r2 = self.minDist(uinterval=(umin, newuMid), vinterval=(newvMid, vmax))
        r3 = self.minDist(uinterval=(newuMid, umax), vinterval=(vmin, newvMid))
        r4 = self.minDist(uinterval=(newuMid, umax), vinterval=(newvMid, vmax))
        results = min([r1, r2, r3, r4], key=lambda x: x[0])
        return results

def curveDistance(bez1,bez2):
    """Find the distance between two curves."""
    c = MinimumCurveDistanceFinder(bez1, bez2)
    dist, t1, t2 = c.minDist()
    return math.sqrt(dist), t1, t2

if __name__ == "__main__":
    bez1 = CubicBezier(
        Point(129, 139), Point(190, 139), Point(201, 364), Point(90, 364)
    )
    bez2 = CubicBezier(
        Point(309, 159), Point(178, 159), Point(215, 408), Point(309, 408)
    )
    bez3 = Line(Point(309, 159), Point(309, 408))

    c = MinimumCurveDistanceFinder(bez1, bez3)
    dist, t1, t2 = c.minDist()
    print(bez1.pointAtTime(t1))
    print(bez3.pointAtTime(t2))
    print(math.sqrt(dist))
    print(c.iterations)



from math import sqrt
try:
    from math import isclose
except ImportError:
    def isclose(a, b):
        return -1e-9 < a - b < 1e-9

def quadraticRoots(a, b, c):
  """Returns real roots of at^2 + bt + c = 0 if 0 < root < 1"""
  roots = []
  if a != 0.0 and b*b-4*a*c > 0.0:
    x = -b / (2 * a)
    y = sqrt(b*b - 4*a*c) / (2*a)
    t1 = x - y
    if 0.0 <= t1 <= 1.0:
      roots.append(t1)
    t2 = x + y
    if 0.0 <= t2 <= 1.0:
      roots.append(t2)
  return roots




import unittest
from beziers.path import BezierPath
from beziers.path.representations.Nodelist import NodelistRepresentation, Node
from beziers.point import Point
from beziers.path.geometricshapes import Circle, Square

class BooleanShapeOperations(unittest.TestCase):
  def drawIt(self, s,c,paths):
    import matplotlib.pyplot as plt
    fig, ax = plt.subplots()
    s.plot(ax,drawNodes=False)
    c.plot(ax)
    for p in paths:
      p.plot(ax,drawNodes=False,color="red")
    plt.show()

  def not_a_test_square_circle_union(self):
    subject = Square(10, origin=Point(5,5))
    clip = Circle(10, origin=Point(15,15))
    paths = subject.union(clip)
    self.drawIt(subject,clip,paths)


import unittest
from beziers.utils.curvefitter import CurveFit
from beziers.path import BezierPath
from beziers.point import Point

class CurveFitterMethods(unittest.TestCase):
  def test_cf1(self):
    nodes = [
      Point(122,102), Point(35,200), Point(228,145), Point(190,46)
    ]
    t = CurveFit._leftTangent(nodes)
    self.assertAlmostEqual(t.x, -0.663890062102)
    self.assertAlmostEqual(t.y,0.747830184896)
    t = CurveFit._rightTangent(nodes)
    self.assertAlmostEqual(t.x,-0.3583470773350791)
    self.assertAlmostEqual(t.y, -0.9335884383203376)

  def test_cf2(self):
    nodes = [
      Point(100,50),
      Point(50,150),
      Point(100,220),
      Point(200,200),
      Point(250,80),
      Point(220,50)
    ]
    path = BezierPath.fromPoints(nodes)
    segs = path.asSegments()
    self.assertEqual(len(segs),2)
    self.assertEqual(segs[0].start, Point(100.0, 50.0))
    self.assertAlmostEqual(segs[0][1].x, 83.333333333)
    self.assertAlmostEqual(segs[0][1].y, 83.333333333)
    self.assertEqual(segs[0].end, Point(50.0, 150.0))
    self.assertAlmostEqual(segs[1][1].x, 50)
    self.assertEqual(segs[1].end, Point(220.0, 50.0))

  def not_a_test_cf3(self):
    import matplotlib.pyplot as plt
    import math
    fig, ax = plt.subplots()
    points = [
      Point(100,50),
      Point(50,150),
      Point(150,250),
      Point(200,220),
      Point(250,80),
      Point(220,50)
    ]
    path = BezierPath.fromPoints(points)
    centroid = path.bounds().centroid
    path.rotate(centroid, math.pi/2)
    path.balance()
    path.plot(ax)
    path.offset(Point(5,5)).plot(ax, color="red")
    path.offset(Point(-5,-5)).plot(ax, color="green")
    plt.show()


class SampleMixin(object):
  def sample(self,samples):
    """Samples a segment or path a given number of times, returning a list of Point objects.
    Remember that for a Bezier path, the points are not guaranteed to be distributed
    at regular intervals along the path. If you want to space your points regularly,
    use the `regularSample` method instead.

.. figure:: sampling.png
   :scale: 50 %
   :alt: sample versus regularSample

   In the figure, the green ticks are generated by `sample`. These are evenly
   distributed by curve time, but because of the curvature of the curve, there
   are concentrations of samples around the tighter parts of the curve.

   The red ticks are generated by `regularSample`, which evenly spaces the
   samples along the length of the curve.

    """
    step = 1.0 / float(samples)
    t = 0.0
    samples = []
    while t <= 1.0:
      samples.append(self.pointAtTime(t))
      t += step
    if t != 1.0:
      samples.append(self.pointAtTime(1))
    return samples

  def regularSample(self,samples):
    """Samples a segment or path a given number of times, returning a list of Point objects,
    but ensuring that the points are regularly distributed along the length
    of the curve. This is an expensive operation because I am a lazy programmer."""

    return [ self.pointAtTime(t) for t in self.regularSampleTValue(samples) ]

  def regularSampleTValue(self,samples):
    """Sometimes you don't want the points, you just want a set of time values (t) which
    represent regular spaced samples along the curve. Use this method to get a list of time
    values instead of Point objects."""
    # Build LUT; could cache it *if* we knew when to invalidate
    lut = []
    length = self.length
    if length == 0: return []
    step = 1.0 / length
    t = 0
    while t <= 1.0:
      lut.append( (t,self.lengthAtTime(t)) ) # Inefficient algorithm but computers are getting faster
      t += step
    desiredLength = 0.0
    rSamples = []
    while desiredLength < length:
      while len(lut) > 0 and lut[0][1] < desiredLength:
        lut.pop(0)
      if len(lut) == 0:
        break
      rSamples.append(lut[0][0])
      desiredLength += length / samples
    if rSamples[-1] != 1.0:
      rSamples.append(1.0)
    return rSamples



from fontTools.pens.basePen import BasePen
from beziers.path.representations.Nodelist import NodelistRepresentation, Node
from beziers.path import BezierPath


class BezierPathCreatingPen(BasePen):
  def __init__(self, *args, **kwargs):
    super(BezierPathCreatingPen, self).__init__(*args, **kwargs)
    self.paths = []
    self.path = BezierPath()
    self.nodeList = []
  def _moveTo(self, p):
    self.nodeList = [Node(p[0], p[1], "move")]
  def _lineTo(self, p):
    self.nodeList.append(Node(p[0], p[1], "line"))
  def _curveToOne(self, p1, p2, p3):
    self.nodeList.append(Node(p1[0], p1[1], "offcurve"))
    self.nodeList.append(Node(p2[0], p2[1], "offcurve"))
    self.nodeList.append(Node(p3[0], p3[1], "curve"))
  def _qCurveToOne(self, p1, p2):
    self.nodeList.append(Node(p1[0], p1[1], "offcurve"))
    self.nodeList.append(Node(p2[0], p2[1], "curve"))
  def _closePath(self):
    self.path.closed = True
    self.path.activeRepresentation = NodelistRepresentation(self.path, self.nodeList)
    self.paths.append(self.path)
    self.path = BezierPath()



from beziers.point import Point
from beziers.path import BezierPath
from beziers.cubicbezier import CubicBezier
from beziers.line import Line
import math

CIRCULAR_SUPERNESS = 4./3.*(math.sqrt(2)-1)

west  = Point(-1,0)
east  = Point(1,0)
north = Point(0,1)
south = Point(0,-1)

def Circle(x_radius, origin=None, superness=CIRCULAR_SUPERNESS):
  """Returns a path representing a circle of given radius. You can specify the
  `origin` as a Point and the `superness` of the circle."""
  return Ellipse(x_radius, x_radius, origin=origin, superness=superness)

def Ellipse(x_radius, y_radius, origin=None, superness=CIRCULAR_SUPERNESS):
  """Returns a path representing an ellipse of given x and y radii.
  You can specify the `origin` as a Point and the `superness` of the ellipse."""
  if not origin:
    origin = Point(0,0)
  w = origin + west * x_radius
  e = origin + east * x_radius
  n = origin + north * y_radius
  s = origin + south * y_radius

  w_n = CubicBezier(w,
                    w + north * y_radius * superness,
                    n + west * x_radius * superness,
                    n)
  n_e = CubicBezier(n,
                    n + east * x_radius * superness,
                    e + north * y_radius * superness,
                    e)
  e_s = CubicBezier(e,
                  e + south * y_radius * superness,
                  s + east * x_radius * superness,
                  s)
  s_w = CubicBezier(s,
                  s + west * x_radius * superness,
                  w + south * y_radius * superness,
                  w)
  return BezierPath.fromSegments([w_n, n_e, e_s, s_w])

def Rectangle(width, height, origin=None):
  """Returns a path representing an rectangle of given width and height.
  You can specify the `origin` as a Point."""
  if not origin:
    origin = Point(0,0)
  tl = origin + west * width / 2.0 + north * height / 2.0
  tr = origin + east * width / 2.0 + north * height / 2.0
  bl = origin + west * width / 2.0 + south * height / 2.0
  br = origin + east * width / 2.0 + south * height / 2.0

  return BezierPath.fromSegments([
    Line(tl,tr), Line(tr,br), Line(br,bl), Line(bl,tl)
  ])

def Square(width, origin=None):
  """Returns a path representing a square of given width.
  You can specify the `origin` as a Point."""
  return Rectangle(width, width, origin=origin)



from beziers.path.representations.Nodelist import Node
from beziers.line import Line
from beziers.cubicbezier import CubicBezier
from beziers.quadraticbezier import QuadraticBezier
from beziers.point import Point
from beziers.utils import isclose

class SegmentRepresentation(object):
  def __init__(self, path, segments=[]):
    self.path = path
    from beziers.path import BezierPath
    assert isinstance(path, BezierPath)
    self.segments = []
    if segments:
      self.segments = segments

  def data(self):
    return self.segments

  def toNodelist(self):
    first = self.segments[0][0]
    nodelist=[]
    if len(self.segments[0]) == 2:
      nodelist.append(Node(first.x, first.y, "line"))
    else:
      nodelist.append(Node(first.x, first.y, "curve"))
    for seg in self.segments:
      if len(seg) == 4:
        nodelist.append(Node(seg[1].x, seg[1].y, "offcurve"))
        nodelist.append(Node(seg[2].x, seg[2].y, "offcurve"))
        nodelist.append(Node(seg[3].x, seg[3].y, "curve"))
      elif len(seg) == 3:
        nodelist.append(Node(seg[1].x, seg[1].y, "offcurve"))
        nodelist.append(Node(seg[2].x, seg[2].y, "curve"))
      else:
        nodelist.append(Node(seg[1].x, seg[1].y, "line"))
    return nodelist

  def appendSegment(self, seg):
    seg = [Point(n[0],n[1]) for n in seg]
    if len(seg) == 2:
      self.segments.append(Line(*seg))
    elif len(seg) == 3:
      self.segments.append(QuadraticBezier(*seg))
    elif len(seg) == 4:
      self.segments.append(CubicBezier(*seg))
    else:
      raise ValueError("Unknown segment type")

  @classmethod
  def fromNodelist(cls, path, nodelist):
    self = SegmentRepresentation(path)
    # Find first oncurve
    firstOncurve = -1
    for ix, n in enumerate(nodelist):
      if n.type != "offcurve":
        firstOncurve = ix
        break
    first = nodelist[firstOncurve]
    seg = [(first.x,first.y)]

    for n in nodelist[firstOncurve+1:]:
      if n.type == "offcurve":
        seg.append((n.x,n.y))
      if n.type == "line" or n.type == "curve":
        seg.append((n.x,n.y))
        self.appendSegment(seg)
        seg = [(n.x,n.y)]
    for n in nodelist[:firstOncurve]:
      if n.type == "offcurve":
        seg.append((n.x,n.y))
      if n.type == "line" or n.type == "curve":
        seg.append((n.x,n.y))
        self.appendSegment(seg)
        seg = [(n.x,n.y)]

    # Closed?
    if self.path.closed:
      if len(seg) == 1 and isclose(seg[-1][0], first.x) and isclose(seg[-1][1], first.y):
        pass
      else:
        seg.append((first.x,first.y))
        self.appendSegment(seg)
    return self



import unittest
from beziers.path.representations.GSPath import GSPathRepresentation
from beziers.path.representations.Segment import SegmentRepresentation
from beziers.path.representations.Nodelist import Node
from beziers.point import Point
from beziers.cubicbezier import CubicBezier
from beziers.line import Line
from beziers.path import BezierPath
from dotmap import DotMap
from beziers.path.geometricshapes import Rectangle


class PathTests(unittest.TestCase):
  # def test_representations(self):
  #   b = DotMap({ "closed": True,
  #   "nodes": [
  #   {"x":385.0, "y":20.0, "type":"offcurve"},
  #   { "x":526.0, "y":79.0, "type":"offcurve"},
  #   { "x":566.0, "y":135.0, "type":"curve"},
  #   { "x":585.0, "y":162.0, "type":"offcurve"},
  #   { "x":566.0, "y":260.0, "type":"offcurve"},
  #   { "x":484.0, "y":281.0, "type":"curve"},
  #   { "x":484.0, "y":407.0, "type":"offcurve"},
  #   { "x":381.0, "y":510.0, "type":"offcurve"},
  #   { "x":255.0, "y":510.0, "type":"curve"},
  #   { "x":26.0, "y":281.0, "type":"line"},
  #   { "x":26.0, "y":155.0, "type":"offcurve"},
  #   { "x":129.0, "y":20.0, "type":"offcurve"},
  #   { "x":255.0, "y":20.0, "type":"curve"}
  #   ]})
  #   path = BezierPath()
  #   path.activeRepresentation = GSPathRepresentation(path,b)
  #   nl = path.asNodelist()
  #   self.assertEqual(len(nl), 13)
  #   self.assertIsInstance(nl[1], Node)
  #   self.assertEqual(nl[1].type,"offcurve")
  #   self.assertAlmostEqual(nl[1].x,526.0)

  #   segs = path.asSegments()
  #   self.assertEqual(len(segs), 5)
  #   self.assertIsInstance(segs[1], CubicBezier)
  #   self.assertIsInstance(segs[2], Line)

  def test_addextremes(self):
    q = CubicBezier(
      Point(42,135), Point(129,242), Point(167,77), Point(65,59)
    )
    ex = q.findExtremes()
    self.assertEqual(len(ex),2)
    path = BezierPath()
    path.closed = False
    path.activeRepresentation = SegmentRepresentation(path,[q])
    path.addExtremes()
    path.balance()
    segs = path.asSegments()
    self.assertEqual(len(segs), 3)
    # import matplotlib.pyplot as plt
    # fig, ax = plt.subplots()
    # path.plot(ax)
    # plt.show()

  def test_overlap(self):
    nodes = [ Node(698.0,413.0,"offcurve"),
      Node(401.0,179.0,"offcurve"),
      Node(401.0,274.0,"curve"),
      Node(401.0,368.0,"offcurve"),
      Node(315.0,445.0,"offcurve"),
      Node(210.0,445.0,"curve"),
      Node(104.0,445.0,"offcurve"),
      Node(18.0,368.0,"offcurve"),
      Node(18.0,274.0,"curve"),
      Node(18.0,179.0,"offcurve"),
      Node(439.0,400.0,"offcurve"),
      Node(533.0,405.0,"curve")
    ]
    p = BezierPath.fromNodelist(nodes)
    p.closed = True
    i = p.getSelfIntersections()
    self.assertEqual(len(i),1)
    self.assertAlmostEqual(i[0].point.x, 377.71521068)

    # import matplotlib.pyplot as plt
    # fig, ax = plt.subplots()
    # p.plot(ax)
    # for n in i:
    #   circle = plt.Circle((n.point.x, n.point.y), 2, fill=True, color="red")
    #   ax.add_artist(circle)
    # plt.show()

    p = BezierPath.fromNodelist([
      Node(310.0,389.0,"line"),
      Node(453.0,222.0,"line"),
      Node(289.0,251.0,"line"),
      Node(447.0,367.0,"line"),
      Node(578.0,222.0,"line"),
      Node(210.0,-8.0,"line"),
    ])

    i = p.getSelfIntersections()
    self.assertEqual(len(i),1)
    self.assertEqual(i[0].point,Point(374.448829525,313.734583702))

  def test_splitatpoints(self):
    p = BezierPath.fromNodelist([
      Node(297.0,86.0,"offcurve"),
      Node(344.0,138.0,"offcurve"),
      Node(344.0,203.0,"curve"),
      Node(344.0,267.0,"offcurve"),
      Node(297.0,319.0,"offcurve"),
      Node(240.0,319.0,"curve"),
      Node(183.0,319.0,"offcurve"),
      Node(136.0,267.0,"offcurve"),
      Node(136.0,203.0,"curve"),
      Node(136.0,138.0,"offcurve"),
      Node(183.0,86.0,"offcurve"),
      Node(240.0,86.0,"curve"),
    ])
    splitlist = []
    for seg in p.asSegments():
      for t in seg.regularSampleTValue(5):
        splitlist.append((seg,t))
    p.splitAtPoints(splitlist)
    self.assertEqual(len(p.asSegments()),24)

  def test_inside(self):
    p = BezierPath.fromNodelist([
      Node(329,320,"line"),
      Node(564,190,"line"),
      Node(622,332,"offcurve"),
      Node(495,471,"offcurve"),
      Node(329,471,"curve"),
      Node(164,471,"offcurve"),
      Node(34,334,"offcurve"),
      Node(93,190,"curve")
    ])
    self.assertTrue(p.pointIsInside(Point(326,423)))
    self.assertFalse(p.pointIsInside(Point(326,123)))
    self.assertFalse(p.pointIsInside(Point(326,251)))
    self.assertTrue(p.pointIsInside(Point(526,251)))
    self.assertTrue(p.pointIsInside(Point(126,251)))

  def test_area(self):
    p = Rectangle(200, 100)
    self.assertEqual(p.area, 200 * 100)
    self.assertEqual(p.signed_area, -200 * 100)
    self.assertEqual(p.direction, -1)
    p.reverse()
    self.assertEqual(p.signed_area, 200 * 100)
    self.assertEqual(p.direction, 1)




from beziers.segment import Segment
from beziers.point import Point
from beziers.utils import isclose

import math
import sys

class Line(Segment):
  """Represents a line segment within a Bezier path."""
  def __init__(self, start, end):
    self.points = [start,end]
    self._orig = None

  def __repr__(self):
    return "L<%s--%s>" % (self.points[0], self.points[1])

  @classmethod
  def fromRepr(klass,text):
    import re
    p = re.compile("^L<(<.*?>)--(<.*?>)>$")
    m = p.match(text)
    return klass(Point.fromRepr(m.group(1)),Point.fromRepr(m.group(2)))

  def pointAtTime(self,t):
    """Returns the point at time t (0->1) along the line."""
    return self.start.lerp(self.end, t)

  # XXX One of these is wrong
  def tangentAtTime(self,t):
    """Returns the tangent at time t (0->1) along the line."""
    return Point.fromAngle(math.atan2(self.end.y-self.start.y,self.end.x-self.start.x))

  def normalAtTime(self,t):
    """Returns the normal at time t (0->1) along the line."""
    return self.tangentAtTime(t).rotated(Point(0,0),math.pi/2)

  def curvatureAtTime(self,t):
    return sys.float_info.epsilon # Avoid divide-by-zero

  def splitAtTime(self,t):
    """Returns two segments, dividing the given segment at a point t (0->1) along the line."""
    return (Line(self.start, self.pointAtTime(t)), Line(self.pointAtTime(t), self.end))

  def _findRoots(self, dimension):
    if dimension == "x":
        t = self[0].x / (self[0].x - self[1].x) if not isclose(self[0].x, self[1].x) else 100.
    elif dimension == "y":
        t = self[0].y / (self[0].y - self[1].y) if not isclose(self[0].y, self[1].y) else 100.
    else:
        raise SyntaxError("Meh")
    if 0. <= t <= 1.:
        return [t]
    return []

  def tOfPoint(self, point, its_on_the_line_i_swear=False):
    """Returns the t (0->1) value of the given point, assuming it lies on the line, or -1 if it does not."""
    # Just find one and hope the other fits
    # point = self.start * (1-t) + self.end * t
    if not isclose(self.end.x, self.start.x):
      t = (point.x - self.start.x) / (self.end.x-self.start.x)
    elif not isclose(self.end.y, self.start.y):
      t = (point.y - self.start.y) / (self.end.y-self.start.y)
    else:
      print("! Line %s is actually a point..." % self)
      return -1
    if its_on_the_line_i_swear or self.pointAtTime(t).distanceFrom(point) < 2e-7: return t
    return -1

  def flatten(self,degree=8):
    return [self]

  @property
  def slope(self):
    v = self[1]-self[0]
    if v.x == 0: return 0
    return v.y / v.x

  @property
  def intercept(self):
    return self[1].y - self.slope * self[1].x

  @property
  def length(self):
    return self[0].distanceFrom(self[1])

  def findExtremes(self):
    return []

  @property
  def area(self):
    return 0.5 * (self[1].x - self[0].x)*(self[0].y + self[1].y)



import unittest
from beziers.quadraticbezier import QuadraticBezier
from beziers.point import Point

class QuadraticMethods(unittest.TestCase):
  def test_split(self):
    # console.log((new Bezier(150,40,80,30,105,150)).split(0.2))
    q = QuadraticBezier(
      Point(150,40), Point(80,30), Point(105,150))
    left, right = q.splitAtTime(0.2)
    self.assertEqual(left[1],Point(136,38))
    self.assertEqual(right[1],Point(85,54))

  def test_extremes(self):
    q = QuadraticBezier(
      Point(70,250), Point(13,187), Point(209,58))
    r = q.findExtremes()
    self.assertEqual(len(r), 1)
    self.assertAlmostEqual(r[0], 0.22529644268774704)

  def test_extremes2(self):
    # console.log((new Bezier(127,242,71,150,210,60)).extrema())
    q = QuadraticBezier(
      Point(127,242), Point(71,150), Point(210,60))
    r = q.findExtremes()
    self.assertEqual(len(r), 1)
    self.assertAlmostEqual(r[0], 0.28717948717948716)

  def test_extremes3(self):
    # console.log((new Bezier(127,242,27,5,210,60)).extrema())
    q = QuadraticBezier(
      Point(127,242), Point(27,5), Point(210,60))
    r = q.findExtremes()
    self.assertEqual(len(r), 2)
    self.assertAlmostEqual(r[0], 0.35335689045936397)
    self.assertAlmostEqual(r[1], 0.8116438356164384)
    # from beziers.path import BezierPath
    # import matplotlib.pyplot as plt
    # path = BezierPath.fromSegments([q])
    # fig, ax = plt.subplots()
    # path.closed = False
    # path.addExtremes()
    # path.plot(ax)
    # plt.show()


  def test_extremes4(self):
    q = QuadraticBezier(
      Point(664,1075),
      Point(732,1167),
      Point(800,1239)
    )
    r = q.findExtremes()
    self.assertEqual(len(r), 0)




import math

class Point(object):
  """A representation of a point within the Beziers world.

  Here are some things you can do with points. You can interpret
  them as vectors, and add them together::

    >>> a = Point(5,5)
    >>> b = Point(10,10)
    >>> a + b
    <15.0,15.0>

  You can multiply them by a scalar to scale them::

    >>> a * 2
    <10.0,10.0>

  You can adjust them::

    >>> a += b
    >>> a
    <15.0,15.0>

  If you're using Python 3, you can abuse operator overloading
  and compute the dot product of two vectors:

    >>> a = Point(5,5)
    >>> b = Point(10,10)
    >>> a @ b
    100.0

"""

  def __init__(self, x,y):
    self.x = float(x)
    self.y = float(y)

  def __repr__(self):
    return "<%s,%s>" % (self.x,self.y)

  @classmethod
  def fromRepr(klass,text):
    import re
    p = re.compile("^<([^,]+),([^>]+)>$")
    m = p.match(text)
    return klass(m.group(1), m.group(2))

  def __eq__(self, other):
    def isclose(a, b, rel_tol=1e-09, abs_tol=0.0):
      return abs(a-b) <= max(rel_tol * max(abs(a), abs(b)), abs_tol)
    return isclose(self.x, other.x) and isclose(self.y, other.y)

  def __hash__(self):
    return  hash(self.x) << 32 ^ hash(self.y)

  def __mul__(self, other):
    """Multiply a point by a scalar."""
    return Point(self.x * other, self.y * other)

  def __div__(self, other):
    return Point(self.x / other, self.y / other)

  def __truediv__(self, other):
    return Point(self.x / other, self.y / other)

  def __add__(self, other):
    return Point(self.x + other.x, self.y + other.y)

  def __sub__(self, other):
    return Point(self.x - other.x, self.y - other.y)

  def __iadd__(self, other):
    self.x += other.x
    self.y += other.y
    return self

  def __isub__(self, other):
    self.x -= other.x
    self.y -= other.y
    return self

  def __matmul__(self,other): # Dot product. Abusing overloading. Sue me.
    return self.dot(other)

  def dot(self, other):
    return self.x * other.x + self.y * other.y

  def clone(self):
    """Clone a point, returning a new object with the same co-ordinates."""
    return Point(self.x,self.y)

  def rounded(self):
    """Return a point with the co-ordinates truncated to integers"""
    return Point(int(self.x),int(self.y))

  def lerp(self, other, t):
    """Interpolate between two points, at time t."""
    return self * (1-t) + other * (t)

  @property
  def squareMagnitude(self):
    """Interpreting this point as a vector, returns the squared magnitude (Euclidean length) of the vector."""
    return self.x*self.x + self.y*self.y

  @property
  def magnitude(self):
    """Interpreting this point as a vector, returns the magnitude (Euclidean length) of the vector."""
    return math.sqrt(self.squareMagnitude)

  def toUnitVector(self):
    """Divides this point by its magnitude, returning a vector of length 1."""
    mag = self.magnitude
    if mag == 0.0: mag = 1.0
    return Point(self.x/mag, self.y/mag)

  @property
  def angle(self):
    """Interpreting this point as a vector, returns the angle in radians of the vector."""
    return math.atan2(self.y,self.x)

  @property
  def slope(self):
    """Returns slope y/x"""
    if self.x == 0: return 0
    return self.y / self.x

  @classmethod
  def fromAngle(self,angle):
    """Given an angle in radians, return a unit vector representing that angle."""
    return Point(math.cos(angle), math.sin(angle)).toUnitVector()

  def rotated(self,around,by):
    """Return a new point found by rotating this point around another point, by an angle given in radians."""
    delta = around - self
    oldangle = delta.angle
    newangle = oldangle + by
    unitvector = Point.fromAngle(newangle)
    new = around - unitvector * delta.magnitude
    return new

  def rotate(self,around,by):
    """Mutate this point by rotating it around another point, by an angle given in radians."""
    new = self.rotated(around, by)
    self.x = new.x
    self.y = new.y

  def squareDistanceFrom(self,other):
    """Returns the squared Euclidean distance between this point and another."""
    return (self.x - other.x) * (self.x - other.x) + (self.y - other.y) * (self.y - other.y)

  def distanceFrom(self,other):
    """Returns the Euclidean distance between this point and another."""
    return math.sqrt(self.squareDistanceFrom(other))

  def transformed(self, transformation):
    m = transformation.matrix
    x, y = self.x, self.y
    a1, a2, b1 = m[0]
    a3, a4, b2 = m[1]
    xPrime = a1 * x + a2 * y + b1
    yPrime = a3 * x + a4 * y + b2
    return Point(xPrime, yPrime)

  def transform(self, transformation):
    new = self.transformed(transformation)
    self.x = new.x
    self.y = new.y



import sys
from beziers.point import Point
from beziers.utils import isclose
from decimal import Decimal

my_epsilon = 2e-7

class Intersection:
  """An object representing an intersection between two segments.
  The location of the intersection on the first segment is accessible
  as `i.seg1`, `i.t1`. The location of the intersection on the second
  segments is `i.seg2`, `i.t2` and `i.point2` respectively.
  You can get a Point object by calling `i.point`."""

  def __init__(self,seg1,t1,seg2,t2):
    self.seg1   = seg1
    self.t1     = t1
    self.point  = seg1.pointAtTime(t1)
    self.seg2   = seg2
    self.t2     = t2

  def __repr__(self):
    return "I<%s t1=%f t2=%f>" % (self.point, self.t1, self.t2)

class IntersectionsMixin:
  # This isn't something we mix into different classes but I'm
  # just putting it here to keep the code tidy.

  def intersections(self, other, limited = True):
    """Returns an array of `Intersection` objects representing the intersections
    between this Segment and another Segment."""
    # Arrange by degree
    if len(other.points) > len(self.points): self,other = other,self
    if len(self.points) == 4 or len(self.points)==3:
      if len(other.points) == 4 or len(other.points)==3:
        inter = self._curve_curve_intersections(other)
      if len(other.points) == 2:
        inter = self._curve_line_intersections(other)
    elif len(self.points) == 2 and len(other.points) == 2:
        inter = self._line_line_intersections(other)
    else:
      raise ValueError("Couldn't work out which intersection function to use")

    def withinRange(t):
      if t < my_epsilon: return False
      if t > 1.0 + my_epsilon: return False
      return True

    if limited:
      return [ i  for i in inter if withinRange(i.t1) and withinRange(i.t2) ]
    else:
      return inter
    raise "Don't know how to compute intersections of a %s and a %s" % (type(self), type(other))

  def _bothPointsAreOnSameSideOfOrigin(self, a,b,c):
    xDiff = (a.x-c.x) * (b.x-c.x)
    yDiff = (a.y-c.y) * (b.y-c.y)
    return not (xDiff <= 0.0 and yDiff <= 0.0)

  def _line_line_intersections(self, other):
    a = self.start
    b = self.end
    c = other.start
    d = other.end
    if isclose(c.x, d.x) and isclose(a.x, b.x): return []
    if isclose(c.y, d.y) and isclose(a.y, b.y): return []
    if c == d or a == b: return []
    if isclose(b.x,a.x):
      x = a.x
      slope34 = ( d.y - c.y) / ( d.x - c.x )
      y = slope34 * ( x - c.x ) + c.y
      p = Point(x,y)
      i = Intersection(self,self.tOfPoint(p), other, other.tOfPoint(p))
      return [ i ]
    if isclose(c.x,d.x):
      x = c.x
      slope12 = ( b.y - a.y) / ( b.x - a.x )
      y = slope12 * ( x - a.x ) + a.y
      p = Point(x,y)
      i = Intersection(self,self.tOfPoint(p), other, other.tOfPoint(p))
      return [ i ]

    slope12 = ( b.y - a.y) / ( b.x - a.x )
    slope34 = ( d.y - c.y) / ( d.x - c.x )
    if abs(slope12 - slope34) < my_epsilon: return [ ]
    x = ( slope12 * a.x - a.y - slope34 * c.x + c.y ) / ( slope12 - slope34 )
    y = slope12 * ( x - a.x ) + a.y
    intersection = Point(x,y)
    if (self._bothPointsAreOnSameSideOfOrigin(intersection, b, a) and self._bothPointsAreOnSameSideOfOrigin(intersection, c, d)):
      return [ Intersection(self,self.tOfPoint(intersection, its_on_the_line_i_swear=True), other, other.tOfPoint(intersection, its_on_the_line_i_swear=True)) ]
    return []

  def _curve_line_intersections_t(self,line):
    t = line.alignmentTransformation()
    c1 = self.transformed(t)
    intersections = c1._findRoots("y")
    return sorted(intersections)

  def _curve_line_intersections(self,line):
    inter = []
    for t in self._curve_line_intersections_t(line):
      inter.append(Intersection(self,t,line,line.tOfPoint(self.pointAtTime(t), its_on_the_line_i_swear=True)))
    return inter

  def _curve_curve_intersections_t(self,other, precision=1e-3):
    assert(len(self.points) > 2 and len(other.points) > 2)
    if not (self.bounds().overlaps(other.bounds())): return []
    if self.bounds().area < precision and other.bounds().area < precision:
      return [ [
      0.5*(self._range[0] + self._range[1]),
      0.5*(other._range[0] + other._range[1]),
     ] ]
    def xmap(v,ts,te): return ts+(te-ts)*v
    c11, c12 = self.splitAtTime(0.5)
    c11._range = [ self._range[0], xmap(0.5,self._range[0],self._range[1])]
    c12._range = [ xmap(0.5,self._range[0],self._range[1]), self._range[1]]
    c21, c22 = other.splitAtTime(0.5)
    c21._range = [ other._range[0], xmap(0.5,other._range[0],other._range[1])]
    c22._range = [xmap(0.5,other._range[0],other._range[1]), other._range[1]]
    assert(c11._range[0] < c11._range[1])
    assert(c12._range[0] < c12._range[1])
    assert(c21._range[0] < c21._range[1])
    assert(c22._range[0] < c22._range[1])

    found = []
    for this in [c11,c12]:
      for that in [c21,c22]:
        if this.bounds().overlaps(that.bounds()):
          found.extend(this._curve_curve_intersections_t(that, precision))
    seen = {}
    numPrecisionDigits = abs(Decimal(str(precision)).as_tuple().exponent)

    def filterSeen(n):
      # use (precision - 1) digits to check whether we have already
      # seen this value
      key = f"%.{numPrecisionDigits - 1}f" % n[0]

      if key in seen: return False
      seen[key] = 1
      return True
    found = filter(filterSeen, found)
    return found

  def _curve_curve_intersections(self,other):
    assert(len(self.points) > 2 and len(other.points) > 2)
    return [Intersection(self,t[0],other,t[1]) for t in self._curve_curve_intersections_t(other)]



import unittest
from beziers.quadraticbezier import QuadraticBezier
from beziers.point import Point
from beziers.boundingbox import BoundingBox

class BBoxMethods(unittest.TestCase):
  def test_overlap(self):
    b1 = BoundingBox()
    b2 = BoundingBox()
    b1.extend(Point(5,5))
    b1.extend(Point(10,10))
    b2.extend(Point(7,7))
    b2.extend(Point(14,14))
    print("%s v %s" % (b1,b2))
    self.assertTrue(b1.overlaps(b2))
    self.assertTrue(b2.overlaps(b1))

  def test_quadratic_bounds(self):
    # console.log((new Bezier(150,40,80,30,105,150)).bbox())
    q = QuadraticBezier(
      Point(150,40), Point(80,30), Point(105,150))
    b = q.bounds()
    self.assertAlmostEqual(b.bl.x,98.42105263157895)
    self.assertAlmostEqual(b.tr.x,150)
    self.assertAlmostEqual(b.bl.y,39.23076923076923)
    self.assertAlmostEqual(b.tr.y,150)



from beziers.point import Point
from beziers.cubicbezier import CubicBezier
import sys
import math

def B0(u):
  return (1.0 - u) * (1.0 - u) * (1.0 - u)

def B1(u):
  return 3 * u * (1.0 - u) * (1.0 - u)

def B2(u):
  return 3 * u * u * (1.0 - u)

def B3(u):
  return u * u * u


class CurveFit:
  # Algorithm lifted from Inkscape

  @classmethod
  def fitCurve(self, data, error, cornerTolerance, maxSegments):
    # We want to uniqify the points but maintaining order
    # (so we can't use a set). An ordered set would be too heavy for this.
    keys = {}
    def filterSeen(x):
      if hash(x) in keys: return False
      keys[hash(x)] = 1
      return True
    data = list(filter(filterSeen, data))
    if len(data) < 2: return
    return self._fitCurve(data, None, None, error, cornerTolerance, maxSegments)

  @classmethod
  def fitLine(self, data, tHat1, tHat2):
    p0, p3 = data[0], data[-1]
    dist = p0.distanceFrom(p3)/3.0

    if tHat1:
      p1 = p0 + tHat1 * dist
    else:
      p1 = ((p0 * 2.0) + p3) / 3.0
    if tHat2:
      p2 = p3 + tHat2 * dist
    else:
      p2 = ((p3 * 2.0) + p0) / 3.0

    return CubicBezier(p0, p1, p2, p3)

  @classmethod
  def estimateBi(self, bez, data, u):
    num = Point(0, 0)
    den = 0.0
    for (coeff, datum) in zip(u, data):
      b0,b1,b2,b3 =  B0(coeff), B1(coeff), B2(coeff), B3(coeff)
      num.x += b1 * (b0 * bez[0].x + b2 * bez[2].x + b3 * bez[3].x - datum.x)
      num.y += b1 * (b0 * bez[0].y + b2 * bez[2].y + b3 * bez[3].y - datum.y)
      den -= b1 * b1

    if den != 0.0:
      bez[1] = num/den
    else:
      bez[1] = bez[0].lerp(bez[3], 1/3.0)

  @classmethod
  def _leftTangent(self, data):
    return (data[1]-data[0]).toUnitVector()

  @classmethod
  def _rightTangent(self, data):
    return (data[-1]-data[-2]).toUnitVector()

  @classmethod
  def centerTangent(self, data, center):
    if data[center + 1] == data[center-1]:
      ret = data[center] - data[center - 1]
      ret.rotate(Point(0,0), math.pi / 2.0)
      return ret.toUnitVector()
    else:
      ret = data[center - 1] - data[center+1]
      return ret.toUnitVector()

  @classmethod
  def leftTangent(self, data, tolerance):
    i = 1
    while True:
      pi = data[i]
      t = pi - data[0]
      distSq = t.__matmul__(t)
      if (tolerance < distSq): return t.toUnitVector()
      i = i + 1
      if i == len(data):
        if distSq == 0: return self._leftTangent(data)
        else: return t.toUnitVector()

  @classmethod
  def rightTangent(self, data, tolerance):
    i = len(data)-2
    while True:
      pi = data[i]
      t = pi - data[-1]
      distSq = t.__matmul__(t)
      if (tolerance < distSq): return t.toUnitVector()
      i = i - 1
      if i == 0:
        if distSq == 0: return self._rightTangent(data)
        else: return t.toUnitVector()

  @classmethod
  def generateBezier(self, data, u, tHat1, tHat2, tolerance_sq):
    est_tHat1 = tHat1
    est_tHat2 = tHat2
    if not est_tHat1:
      est_tHat1 = self.leftTangent(data, tolerance_sq)
    if not est_tHat2:
      est_tHat2 =  self.rightTangent(data, tolerance_sq)
    bez = self.estimateLengths(data, u, est_tHat1, est_tHat2)
    if not tHat1:
      # print("Refining estimate %s" % bez)
      self.estimateBi(bez, data, u)
      # print("Result of estimateBi %s" % bez)
      if bez[1].distanceFrom(bez[0]) > sys.float_info.epsilon:
        est_tHat1 = (bez[1]-bez[0]).toUnitVector()
      # print("tHat1 is now %s" % est_tHat1)
      bez = self.estimateLengths(data, u, est_tHat1, est_tHat2)
    return bez

  @classmethod
  def estimateLengths(self, data, u, tHat1, tHat2):
    C = [ [ 0.0, 0.0], [ 0.0, 0.0 ]]
    X = [ 0.0, 0.0 ]
    # bez = SCBezier()
    for (coeff,datum) in zip(u, data):
      (b0,b1,b2,b3) = (B0(coeff), B1(coeff), B2(coeff), B3(coeff))
      (a1, a2) = (tHat1 * b1, tHat2 * b2)
      C[0][0] += a1.__matmul__(a1)
      C[0][1] += a1.__matmul__(a2)
      C[1][0] = C[0][1]
      C[1][1] += a2.__matmul__(a2)
      s1 = data[0] * (b0+b1)
      shortfall = datum - s1
      shortfall = shortfall - (data[-1] * (b2+b3))
      X[0] += a1.__matmul__(shortfall)
      X[1] += a2.__matmul__(shortfall)

    det_C0_C1 = C[0][0] * C[1][1] - C[1][0] * C[0][1]
    if det_C0_C1 != 0.0:
      det_C0_X  = C[0][0] * X[1]    - C[0][1] * X[0]
      det_X_C1  = X[0]    * C[1][1] - X[1]    * C[0][1]
      alpha_l = det_X_C1 / det_C0_C1
      alpha_r = det_C0_X / det_C0_C1
    else:
      c0 = C[0][0] + C[0][1]
      if c0 != 0:
        alpha_l = X[0] / c0
        alpha_r = X[0] / c0
      else:
        alpha_l = 0.0
        alpha_r = 0.0
    if alpha_l < 1.0e-6 or alpha_r < 1.0e-6:
      alpha_l = data[-1].distanceFrom(data[0]) / 3.0
      alpha_r = alpha_l

    return CubicBezier(data[0],
                    tHat1 * alpha_l + data[0],
                    tHat2 * alpha_r + data[-1],
                    data[-1])

  @classmethod
  def chordLengthParameterize(self, points):
    u = [0.0]
    v = 0
    for i in range(1, len(points)):
      v += points[i].distanceFrom(points[i-1])
      u.append(v)
    return [ n/v for n in u ]

  @classmethod
  def newtonRaphsonFind(self, bez, point, u):
    q0 = bez.pointAtTime(u)
    q1 = bez.derivative().pointAtTime(u)
    q2 = bez.derivative().derivative().pointAtTime(u)
    diff = q0 - point
    numerator = diff.__matmul__(q1)
    denominator = (q1.__matmul__(q1)) + (diff.__matmul__(q2))
    if denominator > 0.0:
      improvedU = u - ( numerator / denominator )
    else:
      if numerator > 0.0:
        improvedU = u * 0.98 - 0.01
      elif numerator < 0.0:
        improvedU = 0.031 + u * 0.98
      else:
        improvedU = u

    if (improvedU < 0): improvedU = 0
    if (improvedU > 1): improvedU = 1
    dist2 = q0.distanceFrom(point)
    proportion = 0.125
    while True:
      proportion += 0.125
      newDistance2 = point.distanceFrom(bez.pointAtTime(improvedU))
      if newDistance2 > dist2:
        if proportion > 1.0:
          improvedU = u
          break
        improvedU = ( ( 1 - proportion ) * improvedU  + proportion * u)
      else:
       break
    return improvedU

  @classmethod
  def reparameterize(self, bez, points, params):
    for i in range(0, len(points)):
      params[i] = self.newtonRaphsonFind(bez, points[i], params[i])

  @classmethod
  def computeHook(self, ffrom, to, parameter, bez, cornerTolerance):
    p = bez.pointAtTime(parameter)
    dist = p.distanceFrom(ffrom.lerp(to, 0.5))
    if dist < cornerTolerance: return 0
    allowed = ffrom.distanceFrom(to) + cornerTolerance
    return dist / allowed

  @classmethod
  def computeMaxError(self, bez, points, params, tolerance, cornerTolerance):
    maxSqDist = 0.0
    maxHookRatio = 0.0
    splitPoint = 0
    snapEnd = 0
    prev = bez[0]
    for i in range(1, len(points)):
      cur = bez.pointAtTime(params[i])
      distSq = cur.squareDistanceFrom(points[i])
      if distSq > maxSqDist:
        maxSqDist = distSq
        splitPoint = i
      hookRatio = self.computeHook(prev, cur, (params[i]+params[i-1])/(2.0), bez, cornerTolerance)
      if maxHookRatio < hookRatio:
        maxHookRatio = hookRatio
        snapEnd = i
      prev = cur
    distRatio = math.sqrt(maxSqDist) / tolerance
    if maxHookRatio <= distRatio:
      return (distRatio, splitPoint)
    else:
      splitPoint = snapEnd-1
      return (-maxHookRatio, splitPoint)

  @classmethod
  def _fitCurve(self, points, tangent1, tangent2, error, cornerTolerance, maxSegments):
    if len(points) == 0: return
    if len(points) == 2:
      return [self.fitLine(points, tangent1, tangent2)]
    maxIterations = 3
    isCorner = False
    u = self.chordLengthParameterize(points)
    if u[-1] == 0.0: return []
    bez = self.generateBezier(points, u, tangent1, tangent2, error)
    self.reparameterize(bez, points, u)
    tolerance = math.sqrt(error + 1e-9)
    (maxErrorRatio, splitPoint) = self.computeMaxError(bez, points, u, tolerance, cornerTolerance)
    if abs(maxErrorRatio) <= 1.0: return [bez]
    if ( 0.0 <= maxErrorRatio and maxErrorRatio <= 3.0 ):
      for _ in range(0, maxIterations+1):
        bez = self.generateBezier(points, u, tangent1, tangent2, error)
        (maxErrorRatio, splitPoint) = self.computeMaxError( bez,  points,  u,  tolerance,  cornerTolerance)
        if abs(maxErrorRatio) <= 1.0: return [bez]

    isCorner = maxErrorRatio < 0
    if isCorner:
      if splitPoint == 0:
        if tangent1 is None:
          splitPoint = splitPoint + 1
        else:
          return self._fitCurve( points, Point(0.0,0.0), tangent2, error, cornerTolerance, maxSegments)

      elif splitPoint == len(points) - 1:
        if tangent2 is None:
          splitPoint = splitPoint - 1
        else:
          return self._fitCurve(points, tangent1, Point(0.0,0.0), error, cornerTolerance, maxSegments)

    if 1 < maxSegments:
      segmentsRemaining = maxSegments - 1
      if isCorner:
        if not (0 < splitPoint and splitPoint < len(points) - 1): return []
        recTHat1 = Point(0.0,0.0)
        recTHat2 = Point(0.0,0.0)
      else:
        recTHat2 = self.centerTangent(points, splitPoint)
        recTHat1 = recTHat2 * -1

      lPoints = points[:splitPoint+1]
      rPoints = points[splitPoint:]
      lbeziers = self._fitCurve(lPoints, tangent1, recTHat2, error, cornerTolerance, segmentsRemaining)
      if lbeziers:
        segmentsRemaining = segmentsRemaining - len(lbeziers)
      rbeziers = self._fitCurve(rPoints, recTHat1, tangent2, error, cornerTolerance, segmentsRemaining)
      return lbeziers + rbeziers
    else:
      return []



from collections import deque

def dequefilter(deck, condition):
    for _ in range(0, len(deck)):
        item = deck.popleft()
        if condition(item):
            deck.append(item)

def bbox_intersections(seta, setb):
	active_a = deque([])
	active_b = deque([])
	instructions = []
	intersections = []
	def add_to(o, bounds, l):
		l.append( (o, bounds) )
		if l == active_a:
			other = active_b
		else:
			other = active_a
		for (o2, bounds2) in other:
			if bounds.overlaps(bounds2):
				intersections.append( (o, o2) )

	def remove_from(o, bounds, l):
		dequefilter(l, lambda i: i[0] != o)

	for a in seta:
		bounds = a.bounds()
		instructions.append((bounds.left, a, bounds, add_to, active_a))
		instructions.append((bounds.right, a, bounds, remove_from, active_a))
	for b in setb:
		bounds = b.bounds()
		instructions.append((bounds.left, b, bounds, add_to, active_b))
		instructions.append((bounds.right, b, bounds, remove_from, active_b))
	instructions = sorted(instructions, key=lambda i:i[0])
	for key, o, bounds, verb, activelist in instructions:
		verb(o, bounds, activelist)
	return intersections

if __name__ == "__main__":
	from beziers.path.geometricshapes import Rectangle
	from beziers.point import Point
	left = [
		Rectangle(100,70,origin=Point(50,50)),
		Rectangle(30,100,origin=Point(80,150)),
		Rectangle(50,50,origin=Point(200,0)),
	]
	right = [
		Rectangle(100,30,origin=Point(120,0)),
		Rectangle(50,50,origin=Point(0,0)),
		Rectangle(100,250,origin=Point(30,200)),
	]

	import matplotlib.pyplot as plt
	fig, ax = plt.subplots()
	for p in left:
		p.clone().plot(ax,color="blue")
	for p in right:
		p.clone().plot(ax,color="red")
	intersections = bbox_intersections(left,right)
	for l,r in intersections:
		[p.plot(ax,fill="green") for p in l.intersection(r)]
	plt.show()



from beziers.segment import Segment
from beziers.line import Line
from beziers.point import Point
from beziers.quadraticbezier import QuadraticBezier
from beziers.utils.arclengthmixin import ArcLengthMixin

import math
from beziers.utils.legendregauss import Tvalues, Cvalues
from beziers.utils import quadraticRoots

class CubicBezier(ArcLengthMixin,Segment):
  def __init__(self, start, c1,c2,end):
    self.points = [start,c1,c2,end]
    self._range = [0,1]

  def __repr__(self):
    return "B<%s-%s-%s-%s>" % (self[0],self[1],self[2],self[3])

  @classmethod
  def fromRepr(klass,text):
    import re
    p = re.compile("^B<(<.*?>)-(<.*?>)-(<.*?>)-(<.*?>)>$")
    m = p.match(text)
    points = [ Point.fromRepr(m.group(t)) for t in range(1,5) ]
    return klass(*points)

  def pointAtTime(self,t):
    """Returns the point at time t (0->1) along the curve."""
    x = (1 - t) * (1 - t) * (1 - t) * self[0].x + 3 * (1 - t) * (1 - t) * t * self[1].x + 3 * (1 - t) * t * t * self[2].x + t * t * t * self[3].x;
    y = (1 - t) * (1 - t) * (1 - t) * self[0].y + 3 * (1 - t) * (1 - t) * t * self[1].y + 3 * (1 - t) * t * t * self[2].y + t * t * t * self[3].y;
    return Point(x,y)

  def tOfPoint(self,p):
    precision = 1.0/50.0
    bestDist = float("inf")
    bestT = -1
    samples = self.regularSampleTValue(50)
    for t in samples:
      dist = self.pointAtTime(t).distanceFrom(p)
      if dist < bestDist:
        bestDist = dist
        bestT = t
    while precision > 1e-5:
      precision = precision / 2
      lower = bestT - precision
      if lower < 0: lower = 0
      upper = bestT + precision
      if upper > 1: upper = 1
      ldist = self.pointAtTime(lower).distanceFrom(p)
      rdist = self.pointAtTime(lower).distanceFrom(p)
      if ldist < bestDist:
        bestT = lower
        bestDist = ldist
      if rdist < bestDist:
        bestT = upper
        bestDist = rdist
    return bestT

  def splitAtTime(self,t):
    """Returns two segments, dividing the given segment at a point t (0->1) along the curve."""
    p4 = self[0].lerp(self[1],t)
    p5 = self[1].lerp(self[2],t)
    p6 = self[2].lerp(self[3],t)
    p7 = p4.lerp(p5,t)
    p8 = p5.lerp(p6,t)
    p9 = p7.lerp(p8,t)
    return (CubicBezier(self[0],p4,p7,p9), CubicBezier(p9,p8,p6,self[3]))

  def join(self,other):
    """Not currently implemented: join two `CubicBezier` together."""
    raise "Not implemented"

  def toQuadratic(self):
    """Not currently implemented: reduce this to a `QuadraticBezier`."""
    raise "Not implemented"

  def derivative(self):
    """Returns a `QuadraticBezier` representing the derivative of this curve."""
    return QuadraticBezier(
      (self[1]-self[0])*3,
      (self[2]-self[1])*3,
      (self[3]-self[2])*3
    )

  def flatten(self, degree=8):
    samples = self.regularSample(self.length/degree)
    ss = []
    for i in range(1,len(samples)):
      l = Line(samples[i-1], samples[i])
      l._orig = self
      ss.append(l)
    return ss

  def _findRoots(self,dimension):
    def cuberoot(v):
      if v<0: return -math.pow(-v,1/3.0)
      return math.pow(v,1/3.0)

    if dimension == "x":
      pa,pb,pc,pd = self[0].x,self[1].x,self[2].x,self[3].x
    elif dimension == "y":
      pa,pb,pc,pd = self[0].y,self[1].y,self[2].y,self[3].y
    else:
      raise SyntaxError("Meh.")

    a = (3*pa - 6*pb + 3*pc)
    b = (-3*pa + 3*pb)
    c = pa
    d = (-pa + 3*pb - 3*pc + pd)
    if d == 0: return []
    a = a/d
    b = b/d
    c = c/d
    p = (3*b - a*a)/3
    p3 = p/3
    q = (2*a*a*a - 9*a*b + 27*c)/27.0
    q2 = q/2
    discriminant = q2*q2 + p3*p3*p3
    if discriminant < 0:
      mp3  = -p/3
      mp33 = mp3*mp3*mp3
      r    = math.sqrt( mp33 )
      t    = -q / (2*r)
      cosphi = max(min(t,1),-1)
      phi  = math.acos(cosphi)
      crtr = cuberoot(r)
      t1   = 2*crtr
      root1 = t1 * math.cos(phi/3) - a/3
      root2 = t1 * math.cos((phi+2*math.pi)/3) - a/3
      root3 = t1 * math.cos((phi+4*math.pi)/3) - a/3
      roots = [root1, root2, root3]
      return sorted([x for x in roots if x >= 0 and x <= 1])

    if discriminant == 0:
      if q2 < 0:
        u1 = cuberoot(-q2)
      else:
        u1 =-cuberoot(q2);
      root1 = 2*u1 - a/3.0;
      root2 = -u1 - a/3.0;
      roots = [root1,root2]
      return sorted([x for x in roots if x >= 0 and x <= 1])

    sd = math.sqrt(discriminant);
    u1 = cuberoot(sd - q2);
    v1 = cuberoot(sd + q2);
    root1 = u1 - v1 - a/3;
    return [x for x in [root1] if x >= 0 and x <= 1]

  def _findDRoots(self):
    d = self.derivative()
    roots = []

    # We have f(t) = w1 (1-t)^2 + 2 w2 (1-t) t + w3 t^2
    # We want f(t) = a t^2 + b^t + c to solve with the quadratic formula
    roots.extend(quadraticRoots(d[0].x - 2*d[1].x + d[2].x, 2 * (d[1].x-d[0].x), d[0].x))
    roots.extend(quadraticRoots(d[0].y - 2*d[1].y + d[2].y, 2 * (d[1].y-d[0].y), d[0].y))
    return roots

  def findExtremes(self, inflections = False):
    """Returns a list of time `t` values for extremes of the curve."""
    r = self._findDRoots()
    if inflections:
      r.extend(self.derivative()._findDRoots())
    r.sort()
    return [ root for root in r if root >= 0.01 and root <= 0.99 ]

  def curvatureAtTime(self,t):
    """Returns the C curvature at time `t`.."""
    d = self.derivative()
    d2 = d.derivative()
    return d.pointAtTime(t).x * d2.pointAtTime(t).y - d.pointAtTime(t).y * d2.pointAtTime(t).x

  @property
  def tunniPoint(self):
    """Returns the Tunni point of this Bezier (the intersection of
    the handles)."""
    h1 = Line(self[0], self[1])
    h2 = Line(self[2], self[3])
    i = h1.intersections(h2, limited = False)
    if len(i)<1: return
    i = i[0].point
    if i.distanceFrom(self[0]) > 5 * self.length:
      return
    else:
      return i

  def balance(self):
    """Perform Tunni balancing on this Bezier."""
    p = self.tunniPoint
    if not p: return
    if self[0].distanceFrom(p) == 0.0:
      fraction1 = 0.43
    else:
      fraction1 = self[0].distanceFrom(self[1]) / self[0].distanceFrom(p)
    if self[3].distanceFrom(p) == 0.0:
      fraction2 = 0.73
    else:
      fraction2 = self[3].distanceFrom(self[2]) / self[3].distanceFrom(p)
    avg = (fraction2 + fraction1) / 2.0
    if avg > 0 and avg < 1:
      self[1] = self[0].lerp(p, avg)
      self[2] = self[3].lerp(p, avg)

  @property
  def hasLoop(self):
    a1 = self[0].x * (self[3].y - self[2].y) + self[0].y * (self[2].x - self[3].x) + self[3].x * self[2].y - self[3].y * self[2].x
    a2 = self[1].x * (self[0].y - self[3].y) + self[1].y * (self[3].x - self[0].x) + self[0].x * self[3].y - self[0].y * self[3].x
    a3 = self[2].x * (self[1].y - self[0].y) + self[2].y * (self[0].x - self[1].x) + self[1].x * self[0].y - self[1].y * self[0].x
    d3 = 3 * a3
    d2 = d3 - a2
    d1 = d2 - a2 + a1
    l = math.sqrt(d1 * d1 + d2 * d2 + d3 * d3)
    s = 0
    if l != 0: s = 1 / l
    d1 *= s
    d2 *= s
    d3 *= s
    d = 3 * d2 * d2 - 4 * d1 * d3
    if d >= 0: return False
    f1 = math.sqrt(-d)
    f2 = 2 * d1
    return ((d2 + f1) / f2, (d2 - f1) / f2)

  @property
  def area(self):
    """Returns the signed rea between the curve and the y-axis"""
    return (10 * (self[3].x*self[3].y - self[0].x*self[0].y) +
            6 * (self[1].x*self[0].y - self[0].x*self[1].y + self[3].x*self[2].y - self[2].x*self[3].y) +
            3 * (self[2].x*self[0].y - self[0].x*self[2].y + self[2].x*self[1].y - self[1].x*self[2].y +
                 self[3].x*self[1].y - self[1].x*self[3].y) +
                 self[3].x*self[0].y - self[0].x*self[3].y) / 20.



from beziers.utils.legendregauss import Tvalues, Cvalues
import math

class ArcLengthMixin:
  @property
  def length(self):
    d = self.derivative()
    z = 0.5
    _sum = 0
    for i in range(0,len(Tvalues)):
      t = z * Tvalues[i] + z
      p = d.pointAtTime(t)
      arc = math.sqrt(p.x * p.x + p.y * p.y)
      _sum += Cvalues[i] * arc
    return _sum * z


import unittest
from beziers.line import Line
from beziers.point import Point

class LineMethods(unittest.TestCase):
  def test_slope(self):
    l = Line(Point(0,10),Point(20,20.4))
    self.assertAlmostEqual(l.slope, 0.52)

  def test_intercept(self):
    l = Line(Point(0,10),Point(20,20.4))
    self.assertAlmostEqual(l.intercept, 10)




import unittest
from beziers.path.geometricshapes import Circle
from beziers.point import Point
from beziers.path import BezierPath


def drawIt(s, c, segs):
    import matplotlib.pyplot as plt

    fig, ax = plt.subplots()
    s.plot(ax, drawNodes=False)
    c.plot(ax)
    for s in segs:
        BezierPath.fromSegments([s]).plot(ax, drawNodes=False, color="red")
    plt.show()


class DistanceMethods(unittest.TestCase):
    def test_distance(self):
        p1 = Circle(50)
        p2 = Circle(50, origin=Point(200, 0))
        d = p1.distanceToPath(p2)
        # drawIt(p1, p2, [d[3], d[4]])
        self.assertAlmostEqual(d[0], 100)

    def test_distance2(self):
        p1 = Circle(50)
        p2 = Circle(50, origin=Point(100, 100))
        d = p1.distanceToPath(p2)
        # drawIt(p1, p2, [d[3], d[4]])
        self.assertAlmostEqual(d[0], 41.4531774254)



import unittest
from beziers.cubicbezier import CubicBezier
from beziers.line import Line
from beziers.point import Point
from beziers.path import BezierPath
from beziers.path.representations.Segment import SegmentRepresentation

class IntersectionMethods(unittest.TestCase):

  def test_line_line(self):
    l1 = Line(Point(310,389), Point(453,222))
    l2 = Line(Point(289,251), Point(447,367))
    # Sanity check
    self.assertEqual(len(l1.intersections(l2)),1)

  def test_cubic_line(self):
    q = CubicBezier(
      Point(100,240), Point(30,60), Point(210,230), Point(160,30))
    l = Line(Point(25,260), Point(230,20))
    path = BezierPath()
    path.closed = False
    path.activeRepresentation = SegmentRepresentation(path,[q])
    i = q.intersections(l)
    self.assertEqual(len(i),3)
    self.assertEqual(i[0].point,q.pointAtTime(0.117517031451))
    self.assertEqual(i[1].point,q.pointAtTime(0.518591792307))
    self.assertEqual(i[2].point,q.pointAtTime(0.867886610031))
    # print q.intersections(l)
    # import matplotlib.pyplot as plt
    # fig, ax = plt.subplots()
    # path.plot(ax)

    # path = BezierPath()
    # path.closed = False
    # path.activeRepresentation = SegmentRepresentation(path,[l])
    # path.plot(ax)

    # for n in q.intersections(l):
    #   circle = plt.Circle((n.point.x, n.point.y), 1, fill=False)
    #   ax.add_artist(circle)

    # plt.show()

  def test_cubic_cubic(self):
    # q1 = Bezier(10,100, 90,30, 40,140, 220,220)
    # q2 = Bezier(5,150, 180,20, 80,250, 210,190)
    # console.log(q1.intersects(q2))
    q1 = CubicBezier(
        Point(10,100), Point(90,30), Point(40,140), Point(220,220)
    )
    q2 = CubicBezier(
        Point(5,150), Point(180,20), Point(80,250), Point(210,190)
    )
    i = q1.intersections(q2)
    # self.assertEqual(len(i),3)
    # self.assertAlmostEqual(i[0].point.x,81.7904225873)
    # self.assertAlmostEqual(i[0].point.y,109.899396337)
    # self.assertAlmostEqual(i[1].point.x,133.186831292)
    # self.assertAlmostEqual(i[1].point.y,167.148173322)
    # self.assertAlmostEqual(i[2].point.x,179.869157678)
    # self.assertAlmostEqual(i[2].point.y,199.661989162)
    import matplotlib.pyplot as plt
    fig, ax = plt.subplots()

    path = BezierPath()
    path.closed = False
    path.activeRepresentation = SegmentRepresentation(path,[q1])
    path.plot(ax)
    path.activeRepresentation = SegmentRepresentation(path,[q2])
    path.plot(ax)

    for n in i:
      circle = plt.Circle((n.point.x, n.point.y), 2, fill=True, color="red")
      ax.add_artist(circle)

    #plt.show()

  def test_cubic_line_2(self):
    s1 = CubicBezier.fromRepr("B<<584.0,126.03783241124995>-<402.0,163.0378324112499>-<220.00000000000003,200.03783241124995>-<38.0,237.03783241124995>>"),
    ray = Line.fromRepr("L<<357.4,-9.99999999999999>--<357.6,250.2692949284206>>")
    assert (s1[0].intersections(ray))

  def test_cubic_line_3(self):
    seg = CubicBezier.fromRepr("B<<320.0,454.0>-<277.0,454.0>-<230.0,439.0>-<189.0,417.0>>")
    ray = Line.fromRepr("L<<254.5,221.5>--<254.5000000000001,887.6681469418963>>")
    assert seg.intersections(ray)


  def test_cubic_line_4(self):
    seg = CubicBezier.fromRepr("B<<315.0,296.0>-<404.0,296.0>-<468.0,251.0>-<468.0,183.0>>")
    ray = Line.fromRepr("L<<330.0000432054082,365.6789020602332>--<330,286>>")
    assert seg.intersections(ray)




import unittest
from beziers.path.representations.GSPath import GSPathRepresentation
from beziers.path.representations.Segment import SegmentRepresentation
from beziers.path.representations.Nodelist import Node
from beziers.point import Point
from beziers.cubicbezier import CubicBezier
from beziers.line import Line
from beziers.path import BezierPath
from dotmap import DotMap

class PathTests(unittest.TestCase):
  def not_a_test_offset(self):
    b = DotMap({ "closed": False,
    "nodes": [
     {"x": 412.0, "y":500.0, "type":"line"},
     {"x": 308.0, "y":665.0, "type":"offcurve"},
     {"x": 163.0, "y":589.0, "type":"offcurve"},
     {"x": 163.0, "y":504.0, "type":"curve"},
     {"x": 163.0, "y":424.0, "type":"offcurve"},
     {"x": 364.0, "y":321.0, "type":"offcurve"},
     {"x": 366.0, "y":216.0, "type":"curve"},
     {"x": 368.0, "y":94.0, "type":"offcurve"},
     {"x": 260.0, "y":54.0, "type":"offcurve"},
     {"x": 124.0, "y":54.0, "type":"curve"}
    ]})
    path = BezierPath()
    path.activeRepresentation = GSPathRepresentation(path,b)
    import matplotlib.pyplot as plt
    fig, ax = plt.subplots()
    path.addExtremes()
    path.plot(ax)
    for n in path.asSegments():
      p = n.tunniPoint
      if p:
        circle = plt.Circle((p.x, p.y), 1, fill=False, color="blue")
        ax.add_artist(circle)
      n.balance()
    path.translate(Point(5,5))
    path.plot(ax, color="red")
    # o1 = path.offset(Point(10,10))
    # o2 = path.offset(Point(-10,-10))
    # o2.reverse()
    # o1.append(o2)
    # o1.plot(ax)
    plt.show()


import unittest
from beziers.point import Point
from beziers.cubicbezier import CubicBezier
from beziers.affinetransformation import AffineTransformation
import math

class AffineTransformationMethods(unittest.TestCase):
  def test_translate(self):
    p = Point(30,70)
    p.rotate(Point(0,0), math.pi/4)
    self.assertEqual(p.x, -28.284271247461906)
    self.assertEqual(p.y, 70.71067811865476)

    p = Point(30,70)
    m = AffineTransformation.rotation(math.pi/4)
    p.transform(m)
    self.assertAlmostEqual(p.x, -28.284271247461906)
    self.assertAlmostEqual(p.y, 70.71067811865476)

    p = Point(0,10)
    m = AffineTransformation.translation(Point(5,-2))
    p.transform(m)
    self.assertEqual(p.x, 5)
    self.assertEqual(p.y, 8)

  def test_scale(self):
    p = Point(4,5)
    m = AffineTransformation.scaling(2)
    p.transform(m)
    self.assertEqual(p.x, 8)
    self.assertEqual(p.y, 10)

    p = Point(4,5)
    m = AffineTransformation.scaling(1.5, -2)
    p.transform(m)
    self.assertEqual(p.x, 6)
    self.assertEqual(p.y, -10)

  def test_multiple_application(self):
    p = Point(10,10)
    m = AffineTransformation()
    m.translate(Point(6,5))
    m.scale(1.5,2)
    p.transform(m)
    self.assertEqual(p.x, 24)
    self.assertEqual(p.y, 30)



from beziers.point import Point

class Node(object):
  def __init__(self, x,y,type):
    self.point = Point(x,y)
    self.type = type

  @property
  def x(self):
    return self.point.x

  @property
  def y(self):
    return self.point.y

  def __repr__(self):
    return "<x=%s y=%s %s>" % (self.x,self.y,self.type)

class NodelistRepresentation(object):
  def __init__(self, path, nl = None):
    self.path = path
    from beziers.path import BezierPath
    assert isinstance(path, BezierPath)
    self.nodes = []
    if nl:
      self.nodes = nl

  def data(self):
    return self.nodes

  def toNodelist(self):
    return self.nodes

  def fromNodelist(self):
    return self



import unittest
from beziers.cubicbezier import CubicBezier
from beziers.point import Point
from beziers.path import BezierPath

class CubicMethods(unittest.TestCase):
  def test_extremes(self):
    q = CubicBezier(
      Point(65,59), Point(194,90), Point(220,260), Point(70,261)
    )
    # console.log(Bezier(65,59, 194,90, 220,260, 70,261).extrema())
    r = q.findExtremes()
    self.assertEqual(len(r), 1)
    self.assertAlmostEqual(r[0], 0.5275787707261016)
    r = q.findExtremes(inflections = True)
    self.assertEqual(len(r), 2)
    self.assertAlmostEqual(r[0], 0.4512987012987013)
    self.assertAlmostEqual(r[1], 0.5275787707261016)


  def test_length(self):
    q = CubicBezier(
      Point(120,160), Point(35,200), Point(220,260), Point(220,40)
    )
    self.assertAlmostEqual(q.length,272.87003168)

  def test_align(self):
    q = CubicBezier(
      Point(120,160), Point(35,200), Point(220,260), Point(220,40)
    )
    s = q.aligned()
    self.assertAlmostEqual(s[0].x,0.0)
    self.assertAlmostEqual(s[0].y,0.0)
    self.assertAlmostEqual(s[1].x,-85.14452515537582)
    self.assertAlmostEqual(s[1].y,-39.69143277919774)
    self.assertAlmostEqual(s[2].x,-12.803687993289572)
    self.assertAlmostEqual(s[2].y,140.84056792618557)
    self.assertAlmostEqual(s[3].x,156.2049935181331)
    self.assertAlmostEqual(s[3].y,0.0)

  def test_curvature(self):
    q = CubicBezier(
      Point(122,102), Point(35,200), Point(228,145), Point(190,46)
    )
    self.assertAlmostEqual(q.curvatureAtTime(0.5),-103450.5)

  def test_loop(self):
    q = CubicBezier(
      Point(171,272), Point(388,249), Point(167,444), Point(388,176)
    )
    self.assertTrue(not q.hasLoop)

    q = CubicBezier(
      Point(171,272), Point(595,249), Point(167,444), Point(388,176)
    )
    roots = q.hasLoop
    p1 = q.pointAtTime(roots[0])
    p2 = q.pointAtTime(roots[1])
    self.assertTrue(q.hasLoop)
    self.assertEqual(p1,p2)